# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""Unit tests"""

import enum
import unittest

import guidata.dataset as gds
from guidata.config import _
from guidata.dataset.conv import update_dataset
from guidata.env import execenv


class Fruit(enum.Enum):
    """Example enumeration"""

    apple = _("An apple")
    banana = _("A banana")
    cherry = _("A cherry")


class InterpolationMethod(gds.LabeledEnum):
    """Example LabeledEnum with invariant keys and translated labels"""

    LINEAR = "linear", _("Linear interpolation")
    SPLINE = "spline", _("Spline interpolation")
    PCHIP = "pchip", _("PCHIP interpolation")
    NEAREST = "nearest"  # Uses key as label


class ProcessingMode(gds.LabeledEnum):
    """Another LabeledEnum example for testing"""

    FAST = "fast_mode", _("Fast processing")
    ACCURATE = "accurate_mode", _("Accurate processing")
    BALANCED = "balanced_mode", _("Balanced processing")


CHOICE_DEFAULT = Fruit.banana


class Parameters(gds.DataSet):
    """Example dataset"""

    float1 = gds.FloatItem("float #1", default=150.0, min=1.0, max=250.0)
    float2 = gds.FloatItem("float #2", default=200.0, min=1.0, max=250.0)
    number = gds.IntItem("number", default=5, min=3, max=20)
    string = gds.StringItem("string", default="default string", help="a string item")
    fruit1 = gds.ChoiceItem("fruit", choices=Fruit, default=CHOICE_DEFAULT)
    fruit2 = gds.ChoiceItem("fruit", choices=Fruit)
    fruit3 = gds.ChoiceItem("fruit", choices=Fruit, default=None)
    # LabeledEnum examples
    interpolation = gds.ChoiceItem(
        "interpolation method",
        choices=InterpolationMethod,
        default=InterpolationMethod.LINEAR,
    )
    processing_mode = gds.ChoiceItem(
        "processing mode", choices=ProcessingMode, default=ProcessingMode.BALANCED
    )


class TestCheck(unittest.TestCase):
    def test_range(self):
        """Test range checking of FloatItem"""
        e = Parameters()
        e.float1 = 150.0
        e.float2 = 400.0
        e.number = 4
        e.fruit3 = CHOICE_DEFAULT  # Avoid None value (error)
        e.interpolation = InterpolationMethod.LINEAR  # Avoid None value (error)
        errors = e.check()
        self.assertEqual(errors, ["float2"])

    def test_typechecking(self):
        """Test type checking of FloatItem"""
        e = Parameters()
        e.string = 400
        e.number = 4.0
        e.fruit3 = CHOICE_DEFAULT  # Avoid None value (error)
        e.interpolation = InterpolationMethod.LINEAR  # Avoid None value (error)
        errors = e.check()
        self.assertEqual(errors, ["number", "string"])

    def test_update(self):
        """Test dataset update"""
        e1 = Parameters()
        e2 = Parameters()
        e1.float1 = 23
        update_dataset(e2, e1)
        self.assertEqual(e2.float1, 23)

    def test_choice_item_default(self):
        """Test choice item default values"""
        e = Parameters()
        # Check default value:
        self.assertEqual(e.fruit1, CHOICE_DEFAULT)
        self.assertIs(e.fruit2, Fruit.apple)  # First choice by default
        self.assertIsNone(e.fruit3)  # Default is None, value is None

    def test_choice_item_set(self):
        """Test choice item setting values"""
        e = Parameters()
        # Test standard values
        e.fruit1 = Fruit.cherry
        e.fruit2 = Fruit.banana
        e.fruit3 = Fruit.apple
        self.assertEqual(e.fruit1, Fruit.cherry)
        self.assertEqual(e.fruit2, Fruit.banana)
        self.assertEqual(e.fruit3, Fruit.apple)
        # Test index assignment
        e.fruit1 = 2
        e.fruit2 = 1
        e.fruit3 = 0
        self.assertEqual(e.fruit1, Fruit.cherry)
        self.assertEqual(e.fruit2, Fruit.banana)
        self.assertEqual(e.fruit3, Fruit.apple)
        # Test names
        e.fruit1 = "cherry"
        e.fruit2 = "banana"
        e.fruit3 = "apple"
        self.assertEqual(e.fruit1, Fruit.cherry)
        self.assertEqual(e.fruit2, Fruit.banana)
        self.assertEqual(e.fruit3, Fruit.apple)

    def test_choice_item_invalid(self):
        """Test choice item invalid values"""
        e = Parameters()
        with self.assertRaises(ValueError):
            e.fruit1 = "invalid"
        with self.assertRaises(ValueError):
            e.fruit2 = 2.3

    def test_labeled_enum_structure(self):
        """Test LabeledEnum structure and properties"""
        # Test that LabeledEnum has correct structure
        self.assertEqual(InterpolationMethod.LINEAR.value, "linear")
        self.assertEqual(InterpolationMethod.LINEAR.label, "Linear interpolation")
        self.assertEqual(str(InterpolationMethod.LINEAR), "Linear interpolation")

        # Test enum without explicit label uses value as label
        self.assertEqual(InterpolationMethod.NEAREST.value, "nearest")
        self.assertEqual(InterpolationMethod.NEAREST.label, "nearest")
        self.assertEqual(str(InterpolationMethod.NEAREST), "nearest")

    def test_labeled_enum_choice_item_default(self):
        """Test LabeledEnum choice item default values"""
        e = Parameters()
        # Check default values
        self.assertEqual(e.interpolation, InterpolationMethod.LINEAR)
        self.assertEqual(e.processing_mode, ProcessingMode.BALANCED)

        # Check that we get back enum members, not raw strings
        self.assertIsInstance(e.interpolation, InterpolationMethod)
        self.assertIsInstance(e.processing_mode, ProcessingMode)

    def test_labeled_enum_choice_item_set(self):
        """Test LabeledEnum choice item setting values with different input types"""
        e = Parameters()

        # Test setting with enum members
        e.interpolation = InterpolationMethod.SPLINE
        self.assertEqual(e.interpolation, InterpolationMethod.SPLINE)

        # Test setting with name strings
        e.interpolation = "PCHIP"
        self.assertEqual(e.interpolation, InterpolationMethod.PCHIP)

        # Test setting with value strings (invariant keys)
        e.interpolation = "linear"
        self.assertEqual(e.interpolation, InterpolationMethod.LINEAR)

        # Test setting with label strings (UI display)
        e.interpolation = "Spline interpolation"
        self.assertEqual(e.interpolation, InterpolationMethod.SPLINE)

        # Test setting with indices
        e.interpolation = 0  # LINEAR
        self.assertEqual(e.interpolation, InterpolationMethod.LINEAR)
        e.interpolation = 1  # SPLINE
        self.assertEqual(e.interpolation, InterpolationMethod.SPLINE)

    def test_labeled_enum_internal_storage(self):
        """Test that LabeledEnum items store the name internally"""
        e = Parameters()
        e.interpolation = InterpolationMethod.SPLINE

        # get_value should return the raw storage key (name)
        choice_item = Parameters.interpolation
        internal_value = choice_item.get_value(e)
        self.assertEqual(internal_value, "SPLINE")

        # But __get__ should return the enum member
        self.assertEqual(e.interpolation, InterpolationMethod.SPLINE)
        self.assertEqual(e.interpolation.value, "spline")
        self.assertEqual(e.interpolation.label, "Spline interpolation")

    def test_labeled_enum_choices_structure(self):
        """Test that LabeledEnum generates correct choices structure"""
        choice_item = Parameters.interpolation
        choices = choice_item.get_prop("data", "choices")

        # Should be [(name, label, None), ...]
        expected_choices = [
            ("LINEAR", "Linear interpolation", None),
            ("SPLINE", "Spline interpolation", None),
            ("PCHIP", "PCHIP interpolation", None),
            ("NEAREST", "nearest", None),  # Uses key as label
        ]
        self.assertEqual(choices, expected_choices)

    def test_labeled_enum_invalid_values(self):
        """Test that invalid LabeledEnum values raise appropriate errors"""
        e = Parameters()

        with self.assertRaises(ValueError) as cm:
            e.interpolation = "invalid_method"
        self.assertIn("Invalid value 'invalid_method'", str(cm.exception))
        self.assertIn("InterpolationMethod", str(cm.exception))

        with self.assertRaises(ValueError):
            e.interpolation = 99  # Invalid index

        with self.assertRaises(ValueError):
            e.interpolation = 3.14  # Invalid type

    def test_labeled_enum_string_interoperability(self):
        """Test seamless interoperability between LabeledEnum members and strings"""
        # Test basic equality
        self.assertTrue(InterpolationMethod.LINEAR == "linear")
        self.assertTrue("linear" == InterpolationMethod.LINEAR)
        self.assertTrue(InterpolationMethod.SPLINE == "spline")
        self.assertFalse(InterpolationMethod.LINEAR == "spline")
        self.assertFalse(InterpolationMethod.LINEAR == "invalid")

        # Test single-value enum (where value == label)
        self.assertTrue(InterpolationMethod.NEAREST == "nearest")
        self.assertTrue("nearest" == InterpolationMethod.NEAREST)

        # Test ProcessingMode enum
        self.assertTrue(ProcessingMode.FAST == "fast_mode")
        self.assertTrue("fast_mode" == ProcessingMode.FAST)

    def test_labeled_enum_function_interoperability(self):
        """Test that functions can accept both enum members and their string values"""

        def process_interpolation(method):
            """Function that should work with both enum and string"""
            if method == InterpolationMethod.LINEAR:
                return "linear_processing"
            elif method == InterpolationMethod.SPLINE:
                return "spline_processing"
            elif method == InterpolationMethod.PCHIP:
                return "pchip_processing"
            elif method == InterpolationMethod.NEAREST:
                return "nearest_processing"
            else:
                return "unknown"

        # Test with enum members
        result = process_interpolation(InterpolationMethod.LINEAR)
        self.assertEqual(result, "linear_processing")
        result = process_interpolation(InterpolationMethod.SPLINE)
        self.assertEqual(result, "spline_processing")
        result = process_interpolation(InterpolationMethod.NEAREST)
        self.assertEqual(result, "nearest_processing")

        # Test with string values - should work identically
        self.assertEqual(process_interpolation("linear"), "linear_processing")
        self.assertEqual(process_interpolation("spline"), "spline_processing")
        self.assertEqual(process_interpolation("nearest"), "nearest_processing")

        # Test ProcessingMode
        def process_mode(mode):
            if mode == ProcessingMode.FAST:
                return "fast_result"
            elif mode == ProcessingMode.ACCURATE:
                return "accurate_result"
            else:
                return "unknown"

        self.assertEqual(process_mode(ProcessingMode.FAST), "fast_result")
        self.assertEqual(process_mode("fast_mode"), "fast_result")

    def test_labeled_enum_set_operations(self):
        """Test that enum members and strings deduplicate correctly in sets"""
        # Test set deduplication
        mixed_set = {InterpolationMethod.LINEAR, "linear", InterpolationMethod.SPLINE}
        self.assertEqual(len(mixed_set), 2)  # Should deduplicate enum and string

        # Test with more complex case
        mode_set = {
            ProcessingMode.FAST,
            "fast_mode",
            ProcessingMode.ACCURATE,
            "accurate_mode",
            ProcessingMode.BALANCED,
        }
        self.assertEqual(len(mode_set), 3)  # Should deduplicate the two FAST entries

        # Test that correct values are in set
        self.assertIn(InterpolationMethod.LINEAR, mixed_set)
        self.assertIn("linear", mixed_set)
        self.assertIn(InterpolationMethod.SPLINE, mixed_set)

    def test_labeled_enum_hash_consistency(self):
        """Test that enum members and their string values have consistent hashing"""
        # Hash should be based on the value, enabling set deduplication
        self.assertEqual(hash(InterpolationMethod.LINEAR), hash("linear"))
        self.assertEqual(hash(InterpolationMethod.SPLINE), hash("spline"))
        self.assertEqual(hash(ProcessingMode.FAST), hash("fast_mode"))

        # Different enum members should have different hashes
        hash1 = hash(InterpolationMethod.LINEAR)
        hash2 = hash(InterpolationMethod.SPLINE)
        self.assertNotEqual(hash1, hash2)

    def test_labeled_enum_inequality_behavior(self):
        """Test inequality behavior with non-matching values"""
        # Test inequality with strings
        self.assertFalse(InterpolationMethod.LINEAR == "wrong_value")
        self.assertFalse("wrong_value" == InterpolationMethod.LINEAR)

        # Test inequality with other types
        self.assertFalse(InterpolationMethod.LINEAR == 42)
        self.assertFalse(InterpolationMethod.LINEAR is None)
        self.assertFalse(InterpolationMethod.LINEAR == [])

        # Test inequality between different enum members
        self.assertFalse(InterpolationMethod.LINEAR == InterpolationMethod.SPLINE)
        self.assertFalse(InterpolationMethod.LINEAR == ProcessingMode.FAST)

    def test_labeled_enum_translated_vs_non_translated(self):
        """Test behavior with translated vs non-translated enum members"""
        # InterpolationMethod.NEAREST is non-translated (single value)
        self.assertTrue(InterpolationMethod.NEAREST == "nearest")
        self.assertEqual(InterpolationMethod.NEAREST.value, "nearest")
        self.assertEqual(InterpolationMethod.NEAREST.label, "nearest")

        # InterpolationMethod.LINEAR is translated (tuple value)
        self.assertTrue(InterpolationMethod.LINEAR == "linear")
        self.assertEqual(InterpolationMethod.LINEAR.value, "linear")
        # Should be translated (different from value)
        self.assertNotEqual(InterpolationMethod.LINEAR.label, "linear")


if __name__ == "__main__":
    unittest.main()
    execenv.print("OK")



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

from __future__ import annotations

import warnings

# Compatibility imports with guidata <= 3.3
from guidata.io.base import BaseIOHandler, GroupContext, WriterMixin
from guidata.io.h5fmt import HDF5Handler, HDF5Reader, HDF5Writer
from guidata.io.inifmt import INIHandler, INIReader, INIWriter
from guidata.io.jsonfmt import JSONHandler, JSONReader, JSONWriter

__all__ = [
    "BaseIOHandler",
    "GroupContext",
    "HDF5Handler",
    "HDF5Reader",
    "HDF5Writer",
    "INIHandler",
    "INIReader",
    "INIWriter",
    "JSONHandler",
    "JSONReader",
    "JSONWriter",
    "WriterMixin",
]

warnings.warn(
    "guidata.dataset.io module is deprecated and will be removed in a future release. "
    "Please use guidata.io instead.",
    DeprecationWarning,
    stacklevel=2,
)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License


"""
Unit tests for arrayeditor.py and its rows/columns insertion/deletion features
"""

# guitest: show

from __future__ import annotations

from typing import Any

import numpy as np

from guidata.env import execenv
from guidata.qthelpers import qt_app_context
from guidata.widgets.arrayeditor import ArrayEditor

DEFAULT_ROW_VALUE = 1
DEFAULT_COL_VALUE = 2
DEFAULT_MASK_VALUE = True
DEFAULT_INS_DEL_COUNT = 3
DEFAULT_INSERTION_INDEX = 1


def _create_3d_array() -> np.ndarray:
    """Creates a 3D numpy array with a single element in the first slice

    Returns:
        A 3D numpy array with a single element in the first slice
    """
    arr_3d = np.zeros((3, 3, 4))
    arr_3d[0, 0, 0] = 1
    arr_3d[0, 0, 1] = 2
    arr_3d[0, 0, 2] = 3
    return arr_3d


REQUIRED_3D_SLICE = [slice(None), slice(None), 0]


BASIC_ARRAYS = (
    ("string array", np.array(["kjrekrjkejr"])),
    ("unicode array", np.array(["ñññéáíó"])),
    (
        "masked array",
        np.ma.array([[1, 0], [1, 0]], mask=[[True, False], [False, False]]),
    ),
    (
        "record array",
        np.zeros(
            (2, 2),
            {
                "names": ("red", "green", "blue"),
                "formats": (np.float32, np.float32, np.float32),
            },
        ),
    ),
    (
        "record array with titles",
        np.array(
            [(0, 0.0), (0, 0.0), (0, 0.0)],
            dtype=[(("title 1", "x"), "|i1"), (("title 2", "y"), ">f4")],
        ),
    ),
    ("bool array", np.array([True, False, True])),
    ("int array", np.array([1, 2, 3], dtype="int8")),
    ("float16 array", np.zeros((5, 5), dtype=np.float16)),
    ("3D array", _create_3d_array()),
)

LABELED_ARRAYS = (
    ("float array", np.random.rand(5, 5), ["a", "b", "c", "d", "e"], None),
    (
        "complex array",
        np.round(np.random.rand(5, 5) * 10) + np.round(np.random.rand(5, 5) * 10) * 1j,
        np.linspace(-12, 12, 5),
        np.linspace(-12, 12, 5),
    ),
)


def insert_rows_and_cols(
    arr: np.ndarray | np.ma.MaskedArray,
    default_row_value: Any = DEFAULT_ROW_VALUE,
    default_col_value: Any = DEFAULT_COL_VALUE,
    index=DEFAULT_INSERTION_INDEX,
    insert_size=DEFAULT_INS_DEL_COUNT,
    default_mask_value=DEFAULT_MASK_VALUE,
) -> np.ndarray | np.ma.MaskedArray:
    """Inserts new rows and columns into a numpy array and returns the result.

    Args:
        arr: numpy array to be edited.
        default_row_value: Default value to insert. Defaults to DEFAULT_ROW_VALUE.
        default_col_value: Default value to insert. Defaults to DEFAULT_COL_VALUE.
        index: index at which to insert. Defaults to DEFAULT_INSERTION_INDEX.
        insert_size: number of rows/cols to insert. Defaults to DEFAULT_INS_DEL_COUNT.
        default_mask_value: Default mask value in case the input array is a MaskedArray.
        Defaults to DEFAULT_MASK_VALUE.

    Returns:
        A numpy array with the new rows and columns inserted.
    """
    if arr.ndim == 1:
        arr.shape = (arr.size, 1)
    (default_np_row_value,) = np.array([default_row_value], dtype=arr.dtype)
    arr_2 = np.insert(arr, (index,) * insert_size, default_np_row_value, 0)

    (default_np_col_value,) = np.array([default_col_value], dtype=arr.dtype)
    arr_3 = np.insert(arr_2, (index,) * insert_size, default_np_col_value, 1)

    if isinstance(arr, np.ma.MaskedArray):
        mask_2 = np.insert(arr.mask, (index,) * insert_size, default_mask_value, 0)
        mask_3 = np.insert(mask_2, (index,) * insert_size, default_mask_value, 1)
        arr_3.mask = mask_3

    return arr_3


MODIFIED_BASIC_ARRAYS = tuple(
    (name, insert_rows_and_cols(arr, 1, 2)) for name, arr in BASIC_ARRAYS
)

MODIFIED_LABELED_ARRAYS = tuple(
    (name, insert_rows_and_cols(arr, 1, 2), labelx, labely)
    for (name, arr, labelx, labely) in LABELED_ARRAYS
)


def launch_arrayeditor_insert(
    data, title="", xlabels=None, ylabels=None
) -> ArrayEditor:
    """Helper routine to launch an arrayeditor and return its result

    Args:
        data: numpy array to be edited.
        title: title of the arrayeditor. Defaults to "".
        xlabels: xlabels to use in the ArrayEditor. Defaults to None.
        ylabels: ylabels to use in the ArrayEditor. Defaults to None.

    Returns:
        An ArrayEditor instance with the given data and labels.
    """
    dlg = ArrayEditor()
    dlg.setup_and_check(
        data, title, xlabels=xlabels, ylabels=ylabels, variable_size=True
    )
    if data.ndim == 3:
        dlg.arraywidget.view.model().set_slice(REQUIRED_3D_SLICE)
    dlg.arraywidget.view.model().insert_row(
        DEFAULT_INSERTION_INDEX, DEFAULT_INS_DEL_COUNT, DEFAULT_ROW_VALUE
    )
    dlg.arraywidget.view.model().insert_column(
        DEFAULT_INSERTION_INDEX, DEFAULT_INS_DEL_COUNT, DEFAULT_COL_VALUE
    )
    return dlg


def launch_arrayeditor_insert_delete(
    data: np.ndarray | np.ma.MaskedArray,
    title="",
    xlabels=None,
    ylabels=None,
) -> ArrayEditor:
    """Creates a new arrayeditor with given data, adds new rows and columns,
    and then deletes them before opening a new dialog box with the result.

    Args:
        data: numpy array to be edited.
        title: title of the arrayeditor. Defaults to "".
        xlabels: xlabels to use in the ArrayEditor. Defaults to None.
        ylabels: ylabels to use in the ArrayEditor. Defaults to None.

    Returns:
        An ArrayEditor instance with the given data and labels.
    """
    dlg = launch_arrayeditor_insert(data, title, xlabels, ylabels)
    dlg.arraywidget.view.model().remove_row(
        DEFAULT_INSERTION_INDEX, DEFAULT_INS_DEL_COUNT
    )
    dlg.arraywidget.view.model().remove_column(
        DEFAULT_INSERTION_INDEX, DEFAULT_INS_DEL_COUNT
    )
    return dlg


def test_arrayeditor() -> None:
    """Test array editor for all supported data types"""
    with qt_app_context():
        for (title, data), (_, awaited_result) in zip(
            BASIC_ARRAYS, MODIFIED_BASIC_ARRAYS
        ):
            new_arr_1 = launch_arrayeditor_insert(data, title).get_value()
            assert (new_arr_1 == awaited_result).all()

            new_arr_2 = launch_arrayeditor_insert_delete(data, title).get_value()
            assert (new_arr_2 == data).all()

        # # TODO: This section can be uncommented when the support for label insertion
        # # alongside new row/values works
        # for (title, data, xlabels, ylabels), (*_, awaited_result) in zip(
        #     LABELED_ARRAYS, MODIFIED_LABELED_ARRAYS
        # ):
        #     new_arr = launch_arrayeditor_insert(data, title, xlabels, ylabels)
        #     # assert (new_arr == awaited_result).all()

        execenv.print("OK")


if __name__ == "__main__":
    test_arrayeditor()
    # pprint(MODIFIED_BASIC_ARRAYS)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Button item test

This item is tested separately from other items because it is special: contrary to
other items, it is purely GUI oriented and has no sense in a non-GUI context.
"""

# guitest: show

from __future__ import annotations

import os.path as osp
import re

from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

import guidata.dataset as gds
from guidata.env import execenv
from guidata.qthelpers import qt_app_context


def information_selectable(parent: QW.QWidget, title: str, text: str) -> None:
    """Show an information message box with selectable text.
    Dialog box is *not* modal."""
    box = QW.QMessageBox(parent)
    box.setIcon(QW.QMessageBox.Information)
    box.setWindowTitle(title)
    if re.search(r"<[a-zA-Z/][^>]*>", text):
        box.setTextFormat(QC.Qt.RichText)
        box.setTextInteractionFlags(QC.Qt.TextBrowserInteraction)
    else:
        box.setTextFormat(QC.Qt.PlainText)
        box.setTextInteractionFlags(
            QC.Qt.TextSelectableByMouse | QC.Qt.TextSelectableByKeyboard
        )
    box.setText(text)
    box.setStandardButtons(QW.QMessageBox.Close)
    box.setDefaultButton(QW.QMessageBox.Close)
    box.setWindowFlags(QC.Qt.Window)  # This is necessary only on non-Windows platforms
    box.setModal(False)
    box.show()


class Parameters(gds.DataSet):
    """
    DataSet test
    The following text is the DataSet 'comment': <br>Plain text or
    <b>rich text<sup>2</sup></b> are both supported,
    as well as special characters (α, β, γ, δ, ...)
    """

    def button_cb(
        dataset: Parameters, item: gds.ButtonItem, value: None, parent: QW.QWidget
    ) -> None:
        """Button callback"""
        execenv.print(f"Button clicked: {dataset}, {item}, {value}, {parent}")
        text = "<br>".join(
            [
                f"<b>Dataset</b>: {'<br>' + '<br>'.join(str(dataset).splitlines())}",
                f"<b>Item</b>: {item}",
                f"<b>Value</b>: {value}",
            ]
        )
        information_selectable(parent, "Button Clicked", text)

    dir = gds.DirectoryItem("Directory", osp.dirname(__file__))
    pattern = gds.StringItem("File pattern", "*.py")
    button = gds.ButtonItem("Help", button_cb, "MessageBoxInformation").set_pos(col=1)
    preview = gds.TextItem("File names preview")


def test_button_item():
    """Test button item"""
    with qt_app_context():
        prm = Parameters()
        execenv.print(prm)
        if prm.edit():
            execenv.print(prm)


if __name__ == "__main__":
    test_button_item()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
DataSetEditGroupBox and DataSetShowGroupBox demo

These group box widgets are intended to be integrated in a GUI application
layout, showing read-only parameter sets or allowing to edit parameter values.
"""

# guitest: show

import numpy as np
from qtpy.QtWidgets import QMainWindow, QSplitter

import guidata.dataset as gds
from guidata.configtools import get_icon
from guidata.dataset.qtwidgets import DataSetEditGroupBox, DataSetShowGroupBox
from guidata.env import execenv
from guidata.qthelpers import (
    add_actions,
    create_action,
    get_std_icon,
    qt_app_context,
    win32_fix_title_bar_background,
)
from guidata.tests.dataset.test_activable_dataset import Parameters
from guidata.widgets import about


class AnotherDataSet(gds.DataSet):
    """
    Example 2
    <b>Simple dataset example</b>
    """

    param0 = gds.ChoiceItem("Choice", ["deazdazk", "aeazee", "87575757"])
    param1 = gds.FloatItem("Foobar 1", default=0, min=0)
    a_group = gds.BeginGroup("A group")
    param2 = gds.FloatItem("Foobar 2", default=0.93)
    param3 = gds.FloatItem("Foobar 3", default=123)
    _a_group = gds.EndGroup("A group")


class ExampleMultiGroupDataSet(gds.DataSet):
    """Example DS with multiple groups"""

    choices = gds.MultipleChoiceItem("Choices", ["a", "b", "c", "d", "e"])
    dictionary = gds.DictItem("Dictionary", {"a": 1, "b": 2, "c": 3})
    param0 = gds.ChoiceItem("Choice", ["deazdazk", "aeazee", "87575757"])
    param1 = gds.FloatItem("Foobar 1", default=0, min=0)
    t_group = gds.BeginTabGroup("T group")
    a_group = gds.BeginGroup("A group")
    param2 = gds.FloatItem("Foobar 2", default=0.93)
    dir1 = gds.DirectoryItem("Directory 1")
    file1 = gds.FileOpenItem("File 1")
    _a_group = gds.EndGroup("A group")
    b_group = gds.BeginGroup("B group")
    param3 = gds.FloatItem("Foobar 3", default=123)
    param4 = gds.BoolItem("Boolean")
    _b_group = gds.EndGroup("B group")
    c_group = gds.BeginGroup("C group")
    param5 = gds.FloatItem("Foobar 4", default=250)
    param6 = gds.DateItem("Date").set_prop("display", format="dd.MM.yyyy")
    param7 = gds.ColorItem("Color")
    _c_group = gds.EndGroup("C group")
    _t_group = gds.EndTabGroup("T group")


class OtherDataSet(gds.DataSet):
    """Another example dataset"""

    group1 = gds.BeginGroup("Group 1")
    title = gds.StringItem("Title", default="Title")
    icon = gds.ChoiceItem(
        "Icon",
        (
            ("python.png", "Python"),
            ("guidata.svg", "guidata"),
            ("settings.png", "Settings"),
        ),
    )
    opacity = gds.FloatItem("Opacity", default=1.0, min=0.1, max=1)
    transform = gds.FloatArrayItem("Transform", default=np.array([1, 2, 3, 4, 5, 6]))
    _group1 = gds.EndGroup("Group 1")
    group2 = gds.BeginGroup("Group 2")
    computed = gds.StringItem("Computed").set_computed(
        lambda ds: f"T: {ds.title}, I: {ds.icon}, O: {ds.opacity}"
    )
    _group2 = gds.EndGroup("Group 2")


class MainWindow(QMainWindow):
    """Main window"""

    def __init__(self):
        QMainWindow.__init__(self)
        win32_fix_title_bar_background(self)
        self.setWindowIcon(get_icon("python.png"))
        self.setWindowTitle("Application example")

        # Instantiate dataset-related widgets:
        self.groupbox1 = DataSetShowGroupBox(
            "Activable dataset", Parameters, comment=""
        )
        self.groupbox2 = DataSetShowGroupBox(
            "Standard dataset", AnotherDataSet, comment=""
        )
        self.groupbox3 = DataSetEditGroupBox(
            "Standard dataset", OtherDataSet, comment=""
        )
        self.groupbox4 = DataSetEditGroupBox(
            "Standard dataset", ExampleMultiGroupDataSet, comment=""
        )
        self.groupbox3.SIG_APPLY_BUTTON_CLICKED.connect(self.update_window)
        self.update_groupboxes()

        splitter = QSplitter(self)
        splitter.addWidget(self.groupbox1)
        splitter.addWidget(self.groupbox2)
        splitter.addWidget(self.groupbox3)
        splitter.addWidget(self.groupbox4)
        self.setCentralWidget(splitter)
        self.setContentsMargins(10, 5, 10, 5)

        # File menu
        file_menu = self.menuBar().addMenu("File")
        quit_action = create_action(
            self,
            "Quit",
            shortcut="Ctrl+Q",
            icon=get_std_icon("DialogCloseButton"),
            tip="Quit application",
            triggered=self.close,
        )
        add_actions(file_menu, (quit_action,))

        # Edit menu
        edit_menu = self.menuBar().addMenu("Edit")
        editparam1_action = create_action(
            self, "Edit dataset 1", triggered=self.edit_dataset1
        )
        editparam2_action = create_action(
            self, "Edit dataset 2", triggered=self.edit_dataset2
        )
        editparam4_action = create_action(
            self, "Edit dataset 4", triggered=self.edit_dataset4
        )
        ro_param4_action = create_action(
            self, "Dataset 4: read-only", toggled=self.ro_dataset4
        )
        add_actions(
            edit_menu,
            (editparam1_action, editparam2_action, editparam4_action, ro_param4_action),
        )

        # ? menu
        help_menu = self.menuBar().addMenu("?")
        about_action = create_action(
            self,
            "About guidata",
            icon=get_std_icon("MessageBoxInformation"),
            triggered=about.show_about_dialog,
        )
        add_actions(help_menu, (about_action,))

    def update_window(self):
        """Update window"""
        dataset = self.groupbox3.dataset
        self.setWindowTitle(dataset.title)
        self.setWindowIcon(get_icon(dataset.icon))
        self.setWindowOpacity(dataset.opacity)

    def update_groupboxes(self):
        """Update groupboxes"""
        self.groupbox1.dataset.set_activable(False)  # This is an activable dataset
        self.groupbox1.get()
        self.groupbox2.get()
        self.groupbox4.get()

    def ro_dataset4(self, readonly: bool):
        self.groupbox4.dataset.set_readonly(readonly)
        self.groupbox4.get()

    def edit_dataset1(self):
        """Edit dataset 1"""
        self.groupbox1.dataset.set_activable(True)  # This is an activable dataset
        if self.groupbox1.dataset.edit(self):
            self.update_groupboxes()

    def edit_dataset2(self):
        """Edit dataset 2"""
        if self.groupbox2.dataset.edit(self):
            self.update_groupboxes()

    def edit_dataset4(self):
        """Edit dataset 4"""
        if self.groupbox4.dataset.edit(self):
            self.update_groupboxes()


def test_editgroupbox():
    """Test editgroupbox"""
    with qt_app_context(exec_loop=True):
        window = MainWindow()
        window.show()
        execenv.print("OK")


if __name__ == "__main__":
    test_editgroupbox()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, Codra, Pierre Raybaut.
# Licensed under the terms of the BSD 3-Clause

"""
genreqs
=======

Module for generating requirements tables in documentation and requirements files
from `pyproject.toml` file.
"""

from __future__ import annotations

import argparse
import os.path as osp
import pathlib
import re

import requests

try:
    import tomllib  # type: ignore # Python 3.11+
except ImportError:
    import tomli as tomllib  # fallback for Python <3.11


def __parse_toml_requirements(pyproject_fname: str) -> dict[str, list[str]] | None:
    """Extract requirements from pyproject.toml file.

    Args:
        pyproject_fname: Path to the pyproject.toml file.

    Returns:
        A tuple containing dictionary of requirements (main project and optional
        dependencies), or None, if pyproject.toml file does not exist
    """
    if not osp.isfile(pyproject_fname):
        return None
    with open(pyproject_fname, "rb") as f:
        data = tomllib.load(f)

    requirements = {}

    # Get main project dependencies
    prj = data["project"]
    requirements["__name"] = prj.get("name", "project")
    requirements["main"] = ["Python" + prj["requires-python"]] + prj["dependencies"]

    # Get optional dependencies
    optional_deps = prj.get("optional-dependencies", {})

    requirements.update(optional_deps)
    return requirements


def __get_pypi_package_info(package: str) -> str:
    """Get package summary from PyPI.

    Args:
        package: Package name

    Returns:
        Package summary or empty string if package not found on PyPI
    """
    if package == "Python":
        return "Python programming language"
    try:
        response = requests.get(f"https://pypi.org/pypi/{package}/json")
    except requests.exceptions.ConnectionError:
        return ""
    if response.status_code != 200:
        return ""
    return response.json()["info"]["summary"]


def __convert_requirements_to_rst_table(reqs: list[str]) -> str:
    """Convert requirements list to RST table.

    Args:
        reqs: Requirements list

    Returns:
        RST table as a string
    """
    requirements = [
        ".. list-table::",
        "    :header-rows: 1",
        "    :align: left",
        "",
        "    * - Name",
        "      - Version",
        "      - Summary",
    ]
    modlist = []
    for req in reqs:
        try:
            # Split by environment marker first (semicolon)
            req_parts = req.split(";", 1)
            req_main = req_parts[0].strip()
            env_marker = req_parts[1].strip() if len(req_parts) > 1 else ""

            # Now split the main requirement by version operators
            mod = re.split(" ?(>=|<=|=|<|>)", req_main)[0]
            ver = req_main[len(mod) :].strip()

            # Add environment marker to version if present
            if env_marker:
                ver = f"{ver} ({env_marker})" if ver else f"({env_marker})"
        except ValueError:
            mod, ver = req, ""
        if mod.lower() in modlist:
            continue
        modlist.append(mod.lower())
        requirements.append("    * - " + mod)
        requirements.append("      - " + ver)
        summary = __get_pypi_package_info(mod)
        requirements.append("      - " + summary)
    return "\n".join(requirements)


def generate_requirements_rst(
    pyproject_fname: str,
    output_directory: str | None = None,
) -> None:
    """Generate install 'requirements.rst' reStructuredText text.
    This reStructuredText text is written in a file which is by default located in
    the `doc` folder of the module.

    Args:
        pyproject_fname: Path to folder containing pyproject.toml or setup.cfg file
        output_directory: Destination path for requirements.rst file (optional).
    """
    requirements = __parse_toml_requirements(pyproject_fname)
    if requirements is None:
        print(f"❌ File not found: {pyproject_fname}")
        return
    name = requirements.pop("__name", "project")
    text = f"""The `{name}` package requires the following Python modules:

{__convert_requirements_to_rst_table(requirements["main"])}"""
    for category, title in (
        ("qt", "GUI support (Qt)"),
        ("dev", "development"),
        ("doc", "building the documentation"),
        ("test", "running test suite"),
    ):
        if category in requirements:
            text += f"""

Optional modules for {title}:

{__convert_requirements_to_rst_table(requirements[category])}"""
    if output_directory is None:
        output_directory = osp.join(osp.dirname(pyproject_fname), "doc")
    with open(osp.join(output_directory, "requirements.rst"), "w") as fdesc:
        fdesc.write(text)
    print(f"✅ Wrote requirements.rst to {output_directory}")


def __extract_exact_requirements(dependencies: list[str]) -> list[str]:
    """Return requirements exactly as written.

    Args:
        dependencies: List of dependency strings from pyproject.toml.

    Returns:
        List of exact requirements.
    """
    return list(dependencies)


def generate_requirements_txt(
    pyproject_fname: str,
    output_fname: str,
    include_optional: bool = False,
) -> None:
    """Process the pyproject.toml file and write requirements to output file.

    Args:
        pyproject_path: Path to the pyproject.toml file.
        output_path: Path to the output requirements file.
        include_optional: If True, include optional dependencies.
    """
    pyproject_path = pathlib.Path(pyproject_fname)
    output_path = pathlib.Path(output_fname)
    if not pyproject_path.exists():
        print(f"❌ File not found: {pyproject_path}")
        return

    with pyproject_path.open("rb") as f:
        data = tomllib.load(f)
    project = data.get("project", {})
    deps = project.get("dependencies", [])

    all_deps = __extract_exact_requirements(deps)

    if include_optional:
        opt_deps = project.get("optional-dependencies", {})
        for extra, deps in opt_deps.items():
            all_deps += __extract_exact_requirements(deps)

    # Remove duplicates and sort
    all_deps = sorted(set(all_deps))

    with output_path.open("w", encoding="utf-8") as f:
        for dep in all_deps:
            f.write(dep + "\n")

    print(f"✅ Wrote {len(all_deps)} requirements to {output_path}")


def main() -> None:
    """Main function to parse arguments and process requirements."""
    parser = argparse.ArgumentParser(
        description="Convert pyproject.toml to documentation or requirements files",
        epilog="Use 'rst' to generate requirements.rst, "
        "'txt' for requirements.txt, "
        "'all' for both formats.",
    )
    subparsers = parser.add_subparsers(dest="command", required=True)

    txt_parser = subparsers.add_parser("txt", help="Generate requirements.txt")
    txt_parser.add_argument(
        "--pyproject",
        default="pyproject.toml",
        help="Path to pyproject.toml (default: pyproject.toml)",
    )
    txt_parser.add_argument(
        "--output",
        default=None,
        help="Output filename (default: requirements.txt)",
    )
    txt_parser.add_argument(
        "--include-optional",
        action="store_true",
        default=True,
        help="Include optional dependencies",
    )

    rst_parser = subparsers.add_parser("rst", help="Generate requirements.rst")
    rst_parser.add_argument(
        "--pyproject",
        default="pyproject.toml",
        help="Path to pyproject.toml (default: pyproject.toml)",
    )
    rst_parser.add_argument(
        "--output",
        default=None,
        help="Output directory for requirements.rst (default: doc)",
    )

    subparsers.add_parser(
        "all", help="Generate both requirements.txt and requirements.rst with defaults"
    )

    args = parser.parse_args()

    if args.command == "txt":
        default_output = "requirements.txt"
        generate_requirements_txt(
            pyproject_fname=osp.abspath(args.pyproject),
            output_fname=osp.abspath(args.output or default_output),
            include_optional=args.include_optional,
        )
    elif args.command == "rst":
        generate_requirements_rst(
            pyproject_fname=osp.abspath(args.pyproject),
            output_directory=osp.abspath(args.output) if args.output else None,
        )
    elif args.command == "all":
        # Generate both requirements.txt and requirements.rst
        generate_requirements_txt(
            pyproject_fname=osp.abspath("pyproject.toml"),
            output_fname=osp.abspath("requirements.txt"),
            include_optional=True,
        )
        generate_requirements_rst(
            pyproject_fname=osp.abspath("pyproject.toml"),
            output_directory=osp.abspath("doc"),
        )
    else:
        parser.print_help()
        raise ValueError("Unknown command: {}".format(args.command))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Generate a dataset class from a function signature

This function is used to generate a dataset from a function signature which has
type annotations. See the example below.
"""

# guitest: show

from __future__ import annotations

import numpy as np

from guidata.dataset import create_dataset_from_func
from guidata.env import execenv


def func_ok(
    a: int, b: float, c: str = "test", d: dict[str, int] = {"x": 1, "y": 3}
) -> None:
    """A function with type annotations"""
    pass


def func_no_type(a, b, c="test"):
    """A function without type annotations"""
    pass


def func_no_default(a: int, b: float, c: str, data: np.ndarray) -> None:
    """A function without default values"""
    pass


def test_dataset_from_func():
    """Test generate dataset class from function"""
    for func in (func_ok, func_no_default):
        execenv.print(func.__name__)
        dataset = create_dataset_from_func(func)
        execenv.print(dataset)
        execenv.print(dataset.create(a=1, b=2.0))
        execenv.print("")
    func = func_no_type
    try:
        create_dataset_from_func(func)
        assert False, "Should have raised a ValueError"
    except ValueError:
        # This is expected
        pass


if __name__ == "__main__":
    test_dataset_from_func()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)
#
# The array editor subpackage was derived from Spyder's arrayeditor.py module
# which is licensed under the terms of the MIT License (see spyder/__init__.py
# for details), copyright © Spyder Project Contributors

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

# ruff: noqa

"""Array editor widget"""

from __future__ import annotations

import io
from typing import Any, Generic, Sequence, cast
from qtpy.compat import getsavefilename

import numpy as np
from qtpy.QtCore import QItemSelection, QItemSelectionRange, QLocale, QPoint, Qt, Slot
from qtpy.QtGui import QCursor, QDoubleValidator, QKeySequence
from qtpy.QtWidgets import (
    QAbstractItemDelegate,
    QApplication,
    QCheckBox,
    QHBoxLayout,
    QInputDialog,
    QItemDelegate,
    QLineEdit,
    QMenu,
    QMessageBox,
    QPushButton,
    QShortcut,
    QTableView,
    QVBoxLayout,
    QWidget,
)

from guidata.config import CONF, _
from guidata.configtools import get_font, get_icon
from guidata.qthelpers import add_actions, create_action, keybinding
from guidata.widgets import about
from guidata.widgets.arrayeditor import utils
from guidata.widgets.arrayeditor.arrayhandler import (
    AnySupportedArray,
    BaseArrayHandler,
    MaskedArrayHandler,
    RecordArrayHandler,
)
from guidata.widgets.arrayeditor.datamodel import (
    ArrayModelType,
    BaseArrayModel,
    DataArrayModel,
    MaskArrayModel,
    MaskedArrayModel,
    RecordArrayModel,
)


class ArrayDelegate(QItemDelegate):
    """Array Editor Item Delegate

    Args:
        dtype: Numpy's dtype of the array to edit
        parent: parent QObject
    """

    def __init__(self, dtype: np.dtype, parent=None) -> None:
        QItemDelegate.__init__(self, parent)
        self.dtype = dtype

    def createEditor(self, parent, option, index) -> QLineEdit | None:
        """Create editor widget"""
        model: BaseArrayModel = index.model()  # type: ignore
        value = model.get_value((index.row(), index.column()))
        if model.get_array().dtype.name == "bool":
            value = not value
            model.setData(index, value)
            return None
        if value is not np.ma.masked:
            editor = QLineEdit(parent)
            editor.setFont(get_font(CONF, "arrayeditor", "font"))
            editor.setAlignment(Qt.AlignmentFlag.AlignCenter)
            if utils.is_number(self.dtype):
                validator = QDoubleValidator(editor)
                validator.setLocale(QLocale("C"))
                editor.setValidator(validator)
            editor.returnPressed.connect(self.commitAndCloseEditor)
            return editor
        return None

    def commitAndCloseEditor(self) -> None:
        """Commit and close editor"""
        editor = self.sender()
        # Avoid a segfault with PyQt5. Variable value won't be changed
        # but at least Spyder won't crash. It seems generated by a bug in sip.
        try:
            self.commitData.emit(editor)
        except AttributeError as e:
            print(e)
            pass
        self.closeEditor.emit(editor, QAbstractItemDelegate.EndEditHint.NoHint)

    def setEditorData(self, editor, index) -> None:
        """Set editor widget's data"""
        if (model := index.model()) is not None and editor is not None:
            text = model.data(index, Qt.ItemDataRole.DisplayRole)
            editor.setText(text)


class DefaultValueDelegate(QItemDelegate):
    """Array Editor Item Delegate

    Args:
        dtype: Numpy's dtype of the array to edit
        parent: parent QObject
    """

    def __init__(self, dtype: np.dtype, parent=None) -> None:
        QItemDelegate.__init__(self, parent)
        self.dtype = dtype
        (self.default_value,) = np.zeros(1, dtype=dtype)

    def createEditor(self, parent, option, index) -> QLineEdit:
        """Create editor widget"""
        editor = QLineEdit(parent)
        editor.setFont(get_font(CONF, "arrayeditor", "font"))
        editor.setAlignment(Qt.AlignmentFlag.AlignCenter)
        if utils.is_number(self.dtype):
            validator = QDoubleValidator(editor)
            validator.setLocale(QLocale("C"))
            editor.setValidator(validator)
        editor.returnPressed.connect(self.commitAndCloseEditor)
        return editor

    def commitAndCloseEditor(self):
        """Commit and close editor"""
        editor = self.sender()
        # Avoid a segfault with PyQt5. Variable value won't be changed
        # but at least Spyder won't crash. It seems generated by a bug in sip.
        try:
            self.commitData.emit(editor)
        except AttributeError as e:
            print(e)
            pass
        self.closeEditor.emit(editor, QAbstractItemDelegate.EndEditHint.NoHint)

    def setEditorData(self, editor, index):
        """Set editor widget's data"""
        if (model := index.model()) is not None and editor is not None:
            text = model.data(index, Qt.ItemDataRole.DisplayRole)
            editor.setText(text)


# TODO: Implement "Paste" (from clipboard) feature
class ArrayView(QTableView, Generic[ArrayModelType]):
    """Array view class

    Args:
        parent: parent QObject
        model: BaseArrayModel to use
        dtype: Numpy's dtype of the array to edit
        shape: Numpy's shape of the array to edit
        variable_size: Flag to indicate if the array dimensions can be modified.
         If a BaseArrayHandler is given as input, the handler should also be in
         readonly mode. Defaults to False.
    """

    def __init__(
        self, parent: QWidget, model: ArrayModelType, dtype, shape, variable_size=False
    ) -> None:
        QTableView.__init__(self, parent)
        self._variable_size = variable_size

        self.setModel(model)
        self.setItemDelegate(ArrayDelegate(dtype, self))
        total_width = 0
        for k in range(self.model().shape[1]):
            total_width += self.columnWidth(k)
        self.viewport().resize(min(total_width, 1024), self.height())
        QShortcut(QKeySequence(QKeySequence.Copy), self, self.copy)
        QShortcut(QKeySequence(QKeySequence.Paste), self, self.paste)
        self.horizontalScrollBar().valueChanged.connect(
            lambda val: self.load_more_data(val, columns=True)
        )
        self.verticalScrollBar().valueChanged.connect(
            lambda val: self.load_more_data(val, rows=True)
        )

        if self._variable_size:
            self._current_row_index = None
            self.vheader_menu = self.setup_header_menu(0)
            vheader = self.verticalHeader()
            vheader.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
            vheader.customContextMenuRequested.connect(self.verticalHeaderContextMenu)

            self._current_col_index = None
            self.hheader_menu = self.setup_header_menu(1)
            hheader = self.horizontalHeader()
            hheader.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
            hheader.customContextMenuRequested.connect(self.horizontalHeaderContextMenu)

        self.cell_menu = self.setup_cell_menu()
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.cellContextMenu)

    def model(self) -> ArrayModelType:
        """Returns the current BaseArrayModel or raises an TypeError

        Raises:
            ValueError: if the model is not a BaseArrayModel

        Returns:
            Current array model
        """
        assert isinstance(model := super().model(), BaseArrayModel)
        return cast(ArrayModelType, model)

    def load_more_data(self, value: int, rows=False, columns=False) -> None:
        """Load more data if needed

        Args:
            value: scrollbar value
            rows: Flag to indicate if rows should be loaded
            columns: Flag to indicate if columns should be loaded
        """
        old_selection = self.selectionModel().selection()
        old_rows_loaded = old_cols_loaded = None

        if rows and value == self.verticalScrollBar().maximum():
            old_rows_loaded = self.model().rows_loaded
            self.model().fetch(rows=rows)

        if columns and value == self.horizontalScrollBar().maximum():
            old_cols_loaded = self.model().cols_loaded
            self.model().fetch(columns=columns)

        if old_rows_loaded is not None or old_cols_loaded is not None:
            # if we've changed anything, update selection
            new_selection = QItemSelection()
            for part in old_selection:
                top = part.top()
                bottom = part.bottom()
                if (
                    old_rows_loaded is not None
                    and top == 0
                    and bottom == (old_rows_loaded - 1)
                ):
                    # complete column selected (so expand it to match updated range)
                    bottom = self.model().rows_loaded - 1
                left = part.left()
                right = part.right()
                if (
                    old_cols_loaded is not None
                    and left == 0
                    and right == (old_cols_loaded - 1)
                ):
                    # compete row selected (so expand it to match updated range)
                    right = self.model().cols_loaded - 1
                top_left = self.model().index(top, left)
                bottom_right = self.model().index(bottom, right)
                part = QItemSelectionRange(top_left, bottom_right)
                new_selection.append(part)
            self.selectionModel().select(
                new_selection, self.selectionModel().ClearAndSelect
            )

    def insert_row(self) -> None:
        """Insert row(s) in the array."""
        if (i := self._current_row_index) is not None:
            (
                i,
                insert_number,
                default_values,
                _new_label,
                valid,
            ) = self.ask_default_inserted_value(i, 0)
            if valid:
                self.model().insert_row(i, insert_number, *default_values)
            self._current_row_index = None

    def remove_row(self) -> None:
        """Remove row(s) in the array"""
        if (i := self._current_row_index) is not None:
            i, remove_number, valid = self.ask_rows_cols_to_remove(i, 0)
            if valid:
                self.model().remove_row(i, remove_number)
            self._current_row_index = None

    def insert_col(self) -> None:
        """Insert column(s) in the array"""
        if (j := self._current_col_index) is not None:
            (
                j,
                insert_number,
                default_value,
                _new_label,
                valid,
            ) = self.ask_default_inserted_value(j, 1)
            if valid:
                self.model().insert_column(j, insert_number, *default_value)
            self._current_col_index = None

    def remove_col(self) -> None:
        """Remove column(s) in the array"""
        if (j := self._current_col_index) is not None:
            j, remove_number, valid = self.ask_rows_cols_to_remove(j, 1)
            if valid:
                self.model().remove_column(j, remove_number)
            self._current_col_index = None

    def ask_default_inserted_value(
        self, index: int, axis: int
    ) -> tuple[int, int, tuple[Any, ...], Any | None, bool]:
        """Create and open a new dialog with a form to input insertion parameters

        Args:
            index: default insertion index (choice made when clicking)
            axis: insertion axis (row (0) or column (1))

        Returns:
            User inputs
        """
        InsertionDataSet = self.model().get_insertion_dataset(index, axis)
        title = (
            _("Row(s) insertion")
            if axis == 0
            else _("Column(s) insertion")
            if axis == 1
            else ""
        )
        insertion_dataset = InsertionDataSet(
            title=title,
            icon="insert.png",
        )

        is_ok = insertion_dataset.edit()
        index_: int = insertion_dataset.index_field
        max_index = (
            self.model()
            .get_array()
            .shape[self.model().correct_ndim_axis_for_current_slice(axis)]
        )
        index_ = max_index if index_ == -1 else index_
        return (
            index_,
            insertion_dataset.insert_number,
            insertion_dataset.get_values_to_insert(),
            insertion_dataset.new_label,
            is_ok,
        )  # type: ignore

    def ask_rows_cols_to_remove(self, index: int, axis: int) -> tuple[int, int, bool]:
        """Create and open a new dialog with a form to input deletion parameters

        Args:
            index: default deletion index (choice made when clicking)
            axis: deletion axis (row (0) or column (1))

        Returns:
            User inputs
        """
        DeletionDataSet = self.model().get_deletion_dataset(index, axis)
        title = (
            _("Row(s) deletion")
            if axis == 0
            else _("Column(s) deletion")
            if axis == 1
            else ""
        )
        deletion_dataset = DeletionDataSet(
            title=title,
            icon="delete.png",
        )
        is_ok = deletion_dataset.edit()
        index_ = deletion_dataset.index_field
        max_index = (
            self.model()
            .get_array()
            .shape[self.model().correct_ndim_axis_for_current_slice(axis)]
        )
        number_to_del = min(deletion_dataset.remove_number, max_index - index_)
        return index_, number_to_del, is_ok  # type: ignore

    def resize_to_contents(self) -> None:
        """Resize cells to contents"""
        QApplication.setOverrideCursor(QCursor(Qt.CursorShape.WaitCursor))
        self.resizeColumnsToContents()
        self.model().fetch(columns=True)
        self.resizeColumnsToContents()
        QApplication.restoreOverrideCursor()

    def setup_cell_menu(self) -> QMenu:
        """Setup context menu

        Returns:
            New QMenu object
        """
        self.copy_action = create_action(
            self,
            _("Copy"),
            shortcut=keybinding("Copy"),
            icon=get_icon("editcopy.png"),
            triggered=self.copy,
            context=Qt.ShortcutContext.WidgetShortcut,
        )
        self.paste_action = create_action(
            self,
            _("Paste"),
            shortcut=keybinding("Paste"),
            icon=get_icon("editpaste.png"),
            triggered=self.paste,
            context=Qt.ShortcutContext.WidgetShortcut,
        )
        self.paste_action.setDisabled(self.model().readonly)
        about_action = create_action(
            self,
            _("About..."),
            icon=get_icon("guidata.svg"),
            triggered=about.show_about_dialog,
        )
        if self._variable_size:
            insert_row_action = create_action(
                self,
                title=_("Insert row(s)"),
                icon=get_icon("insert.png"),
                triggered=self.insert_row,
            )
            insert_col_action = create_action(
                self,
                title=_("Insert column(s)"),
                icon=get_icon("insert.png"),
                triggered=self.insert_col,
            )
            remove_row_action = create_action(
                self,
                title=_("Remove row(s)"),
                icon=get_icon("delete.png"),
                triggered=self.remove_row,
            )
            remove_col_action = create_action(
                self,
                title=_("Remove column(s)"),
                icon=get_icon("delete.png"),
                triggered=self.remove_col,
            )
            actions = (
                self.copy_action,
                self.paste_action,
                None,
                insert_row_action,
                insert_col_action,
                None,
                remove_row_action,
                remove_col_action,
                None,
                about_action,
            )
        else:
            actions = (
                self.copy_action,
                self.paste_action,
                None,
                about_action,
            )
        menu = QMenu(self)
        add_actions(menu, actions)
        return menu

    def setup_header_menu(self, axis: int) -> QMenu:
        """Creates and return a contextual menu for a header depending on input axis.

        Args:
            axis: 0 for rows (vertical header), 1 for columns (horizontal header)

        Returns:
            New QMenu object for the given header
        """
        action_args = {
            0: (
                (_("Insert row(s)"), self.insert_row),
                (_("Remove row(s)"), self.remove_row),
            ),
            1: (
                (_("Insert column(s)"), self.insert_col),
                (_("Remove column(s)"), self.remove_col),
            ),
        }[axis]
        insert_action = create_action(
            self,
            title=action_args[0][0],
            icon=get_icon("insert.png"),
            triggered=action_args[0][1],
        )
        remove_action = create_action(
            self,
            title=action_args[1][0],
            icon=get_icon("delete.png"),
            triggered=action_args[1][1],
        )
        actions = (
            insert_action,
            None,
            remove_action,
        )
        menu = QMenu(self)
        add_actions(menu, actions)
        return menu

    def verticalHeaderContextMenu(self, pos: QPoint) -> None:
        """Reimplement Qt method"""
        vheader = self.verticalHeader()
        self._current_row_index = vheader.logicalIndexAt(pos)
        self.vheader_menu.popup(vheader.mapToGlobal(pos))

    def horizontalHeaderContextMenu(self, pos: QPoint) -> None:
        """Reimplement Qt method"""
        hheader = self.horizontalHeader()
        self._current_col_index = hheader.logicalIndexAt(pos)
        self.hheader_menu.popup(hheader.mapToGlobal(pos))

    def cellContextMenu(self, pos: QPoint) -> None:
        """Reimplement Qt method"""
        try:
            selected_index = self.selectedIndexes()[0]
            self._current_row_index, self._current_col_index = (
                selected_index.row(),
                selected_index.column(),
            )
        except IndexError:  # click outside of cells
            self._current_row_index, self._current_col_index = (
                self.model().total_rows,
                self.model().total_cols,
            )  # we get the index of the last array element to insert after the last row/column

        self.cell_menu.popup(self.viewport().mapToGlobal(pos))

    def keyPressEvent(self, event) -> None:
        """Reimplement Qt method"""
        if event == QKeySequence.Copy:
            self.copy()
        else:
            QTableView.keyPressEvent(self, event)

    def _sel_to_text(self, cell_range: list[QItemSelectionRange]) -> str | None:
        """Copy an array portion to a unicode string

        Args:
            cell_range: list of QItemSelectionRange objects

        Returns:
            String representation of the selected array portion, or None if
             the selection is empty
        """
        if not cell_range:
            return None
        model = self.model()
        row_min, row_max, col_min, col_max = utils.get_idx_rect(cell_range)
        if col_min == 0 and col_max == (model.cols_loaded - 1):
            # we've selected a whole column. It isn't possible to
            # select only the first part of a column without loading more,
            # so we can treat it as intentional and copy the whole thing
            col_max = model.total_cols - 1
        if row_min == 0 and row_max == (model.rows_loaded - 1):
            row_max = model.total_rows - 1

        model.apply_changes()
        _data = model.get_array()  # TODO check if this should apply changes or not
        output = io.BytesIO()

        try:
            np.savetxt(
                output,
                _data[row_min : row_max + 1, col_min : col_max + 1],
                delimiter="\t",
                fmt=model.get_format(),
            )
        except BaseException:
            QMessageBox.warning(
                self,
                _("Warning"),
                _("It was not possible to copy values for this array"),
            )
            return None
        contents = output.getvalue().decode("utf-8")
        output.close()
        return contents

    @Slot()
    def copy(self) -> None:
        """Copy text to clipboard"""
        cliptxt = self._sel_to_text(self.selectedIndexes())
        clipboard = QApplication.clipboard()
        clipboard.setText(cliptxt)

    @Slot()
    def paste(self) -> None:
        """Paste text from clipboard"""
        cliptxt = QApplication.clipboard().text()
        if not cliptxt:
            return
        try:
            data = np.genfromtxt(io.StringIO(cliptxt), delimiter="\t")
        except ValueError:
            data = np.array([])
        if data.size == 0:
            QMessageBox.warning(
                self,
                _("Warning"),
                _("It was not possible to paste values for this array"),
            )
            return

        model = self.model()

        # Determine where to paste
        start_row = self.currentIndex().row()
        start_col = self.currentIndex().column()

        # Iterate and paste each value
        data = np.array(data, dtype=model.get_array().dtype)
        if data.ndim in (0, 1):
            data = data.reshape((-1, 1))
        for i in range(data.shape[0]):
            for j in range(data.shape[1]):
                if (start_row + i < model.total_rows) and (
                    start_col + j < model.total_cols
                ):
                    idx = model.index(start_row + i, start_col + j)
                    model.setData(idx, str(data[i, j]))

        model.dataChanged.emit(
            model.index(start_row, start_col),
            model.index(start_row + data.shape[0] - 1, start_col + data.shape[1] - 1),
        )


class BaseArrayEditorWidget(QWidget):
    """Base ArrayEditdorWidget class. Used to wrap handle n-dimensional normal Numpy's
    ndarray.

    Args:
        parent: parent QObject
        data: Numpy's ndarray or BaseArrayHandler to use.
        readonly: Flag for readonly mode. Defaults to False.
        xlabels: labels for the columns (header). Defaults to None.
        ylabels: labels for the rows (header). Defaults to None.
        variable_size: Flag to indicate if the array dimensions can be modified.
         If a BaseArrayHandler is given as input, the handler should also be in
         readonly mode Defaults to False.
        current_slice: slice of the same dimension as the Numpy ndarray that will.
         Defaults to None
    """

    def __init__(
        self,
        parent,
        data: np.ndarray | BaseArrayHandler,
        readonly=False,
        xlabels=None,
        ylabels=None,
        variable_size=False,
        current_slice: Sequence[slice | int] | None = None,
    ) -> None:
        QWidget.__init__(self, parent)

        self._variable_size = variable_size and not readonly
        self._data: BaseArrayHandler | MaskedArrayHandler
        self._init_handler(data)
        self._init_model(xlabels, ylabels, readonly, current_slice=current_slice)

        format = utils.SUPPORTED_FORMATS.get(self.model.get_array().dtype.name, "%s")
        self.model.set_format(format)
        self.view = ArrayView(
            self,
            self.model,
            self.model.get_array().dtype,
            self._data.shape,
            self._variable_size,
        )

        btn_layout = QHBoxLayout()
        btn_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        btn = QPushButton(get_icon("format.svg"), _("Format"), self)
        # disable format button for int type
        btn.setEnabled(utils.is_float(self._data.dtype))
        btn_layout.addWidget(btn)
        btn.clicked.connect(self.change_format)
        btn = QPushButton(get_icon("resize.svg"), _("Resize"), self)
        btn_layout.addWidget(btn)
        btn.clicked.connect(self.view.resize_to_contents)
        bgcolor = QCheckBox(_("Background color"))
        bgcolor.setChecked(self.model.bgcolor_enabled)
        bgcolor.setEnabled(self.model.bgcolor_enabled)
        bgcolor.stateChanged.connect(self.model.bgcolor)
        btn_layout.addWidget(bgcolor)
        btn_layout.addStretch(1)
        btn = QPushButton(get_icon("copy_all.svg"), _("Copy all"), self)
        btn.setToolTip(_("Copy all array data to clipboard"))
        btn.clicked.connect(self.copy_all_to_clipboard)
        btn_layout.addWidget(btn)
        btn = QPushButton(get_icon("export.svg"), _("Export"), self)
        btn.setToolTip(_("Export array to a file"))
        btn_layout.addWidget(btn)
        btn.clicked.connect(self.export_array)

        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.addLayout(btn_layout)
        self.setLayout(layout)

    def _init_handler(
        self, data: AnySupportedArray | BaseArrayHandler[AnySupportedArray]
    ) -> None:
        """Initializes and set the instance handler to use

        Args:
            data: Numpy's ndarray or BaseArrayHandler to use.
        """
        if isinstance(data, np.ndarray):
            self._data = BaseArrayHandler[AnySupportedArray](data, self._variable_size)
        elif isinstance(data, BaseArrayHandler):
            self._data = data
        else:
            raise TypeError(
                "Given data must be of type np.ndarray or BaseArrayHandler, "
                f"not {type(data)}"
            )

    def _init_model(
        self,
        xlabels: Sequence[str] | None,
        ylabels: Sequence[str] | None,
        readonly: bool,
        current_slice: Sequence[slice | int] | None = None,
    ) -> None:
        """Initializes and set the instance model to use

        Args:
            xlabels: labels for the columns (header). Defaults to None.
            ylabels: labels for the rows (header). Defaults to None.
            readonly: Flag for readonly mode. Defaults to False.
            current_slice: slice of the same dimension as the Numpy ndarray that will.
        """
        self.model = BaseArrayModel(
            self._data,
            xlabels=xlabels,
            ylabels=ylabels,
            readonly=readonly,
            parent=self,
            current_slice=current_slice,
        )

    def accept_changes(self) -> None:
        """Accept changes"""
        self.model.apply_changes()

    def reject_changes(self) -> None:
        """Reject changes"""
        self.model.clear_changes()

    def change_format(self) -> None:
        """Change display format"""
        format, valid = QInputDialog.getText(
            self,
            _("Format"),
            _("Float formatting"),
            QLineEdit.Normal,
            self.model.get_format(),
        )
        if valid:
            format = str(format)
            try:
                format % 1.1
            except BaseException:
                QMessageBox.critical(
                    self, _("Error"), _("Format (%s) is incorrect") % format
                )
                return
            self.model.set_format(format)

    def copy_all_to_clipboard(self) -> None:
        """Copy all array data, including headers, to clipboard"""
        data = self.model.get_array()
        xlabels = self.model.xlabels
        ylabels = self.model.ylabels

        output = io.StringIO()
        c0sep = "" if ylabels is None else "\t"

        # Write column headers
        if xlabels:
            output.write(c0sep + "\t".join(xlabels) + "\n")

        # Write rows with row headers
        for i, row in enumerate(data):
            label = ylabels[i] if ylabels else ""
            output.write(label + c0sep + "\t".join(map(str, row)) + "\n")

        cliptxt = output.getvalue()
        output.close()

        if cliptxt:
            clipboard = QApplication.clipboard()
            clipboard.setText(cliptxt)

    def export_array(self) -> None:
        """Export the array to a CSV file with UTF-8 BOM."""
        filename, _selfilter = getsavefilename(
            self, _("Export array"), "", _("CSV Files") + " (*.csv)"
        )
        if not filename:
            return

        data = self.model.get_array()
        xlabels = self.model.xlabels
        ylabels = self.model.ylabels

        with open(filename, "w", encoding="utf-8-sig") as file:
            c0sep = "" if ylabels is None else "\t"

            # Write column headers
            if xlabels:
                file.write(c0sep + "\t".join(xlabels) + "\n")

            # Write rows
            for i, row in enumerate(data):
                label = ylabels[i] if ylabels else ""
                file.write(label + c0sep + "\t".join(map(str, row)) + "\n")

        if not (xlabels or ylabels):
            # If no labels, overwrite file with simple array
            np.savetxt(
                filename,
                data,
                delimiter="\t",
                fmt=self.model.get_format(),
                encoding="utf-8-sig",
            )


class MaskedArrayEditorWidget(BaseArrayEditorWidget):
    """Same as BaseArrayWidgetEditorWidget but specifically handles MaskedArrayHandler
    and MaskedArrayModel. Specifically the masked data.

    Args:
        parent: parent QObject
        data: Numpy's ndarray or BaseArrayHandler to use.
        readonly: Flag for readonly mode. Defaults to False.
        xlabels: labels for the columns (header). Defaults to None.
        ylabels: labels for the rows (header). Defaults to None.
        variable_size: Flag to indicate if the array dimensions can be modified.
         If a BaseArrayHandler is given as input, the handler should also be in
         readonly mode Defaults to False.
        current_slice: slice of the same dimension as the Numpy ndarray that will.
         Defaults to None
    """

    # _data: MaskedArrayHandler

    def _init_handler(self, data: np.ma.MaskedArray | MaskedArrayHandler) -> None:
        """Initializes and set the instance handler to use

        Args:
            data: Numpy's MaskedArray or MaskedArrayHandler to use.
        """
        if isinstance(data, np.ma.MaskedArray):
            self._data = MaskedArrayHandler(data, self._variable_size)
        elif isinstance(data, MaskedArrayHandler):
            self._data = data
        else:
            raise TypeError(
                "Given data must be of type np.ma.MaskedArray or "
                "MaskedArrayHandler, not {type(data)}"
            )

    def _init_model(
        self,
        xlabels: Sequence[str] | None,
        ylabels: Sequence[str] | None,
        readonly: bool,
        current_slice: Sequence[slice | int] | None = None,
    ) -> None:
        """Initializes and set the instance model to use

        Args:
            xlabels: labels for the columns (header). Defaults to None.
            ylabels: labels for the rows (header). Defaults to None.
            readonly: Flag for readonly mode. Defaults to False.
            current_slice: slice of the same dimension as the Numpy ndarray that will.
        """
        assert isinstance(self._data, MaskedArrayHandler)
        self.model = MaskedArrayModel(
            self._data,
            xlabels=xlabels,
            ylabels=ylabels,
            readonly=readonly,
            parent=self,
            current_slice=current_slice,
        )


class MaskArrayEditorWidget(MaskedArrayEditorWidget):
    """Same as BaseArrayWidgetEditorWidget but specifically handles MaskedArrayHandler
    and MaskArrayModel. Specifically the boolean mask.

    Args:
        parent: parent QObject
        data: Numpy's ndarray or BaseArrayHandler to use.
        readonly: Flag for readonly mode. Defaults to False.
        xlabels: labels for the columns (header). Defaults to None.
        ylabels: labels for the rows (header). Defaults to None.
        variable_size: Flag to indicate if the array dimensions can be modified.
         If a BaseArrayHandler is given as input, the handler should also be in
         readonly mode Defaults to False.
        current_slice: slice of the same dimension as the Numpy ndarray that will.
         Defaults to None
    """

    # _data: MaskedArrayHandler

    def _init_model(
        self,
        xlabels: Sequence[str] | None,
        ylabels: Sequence[str] | None,
        readonly: bool,
        current_slice: Sequence[slice | int] | None = None,
    ) -> None:
        """Initializes and set the instance model to use

        Args:
            xlabels: labels for the columns (header). Defaults to None.
            ylabels: labels for the rows (header). Defaults to None.
            readonly: Flag for readonly mode. Defaults to False.
            current_slice: slice of the same dimension as the Numpy ndarray that will.
        """
        assert isinstance(self._data, MaskedArrayHandler)
        self.model = MaskArrayModel(
            self._data,
            xlabels=xlabels,
            ylabels=ylabels,
            readonly=readonly,
            parent=self,
            current_slice=current_slice,
        )


class DataArrayEditorWidget(MaskedArrayEditorWidget):
    """Same as BaseArrayWidgetEditorWidget but specifically handles MaskedArrayHandler
    and DataArrayModel. Specifically the raw unmasked data.

    Args:
        parent: parent QObject
        data: Numpy's ndarray or BaseArrayHandler to use.
        readonly: Flag for readonly mode. Defaults to False.
        xlabels: labels for the columns (header). Defaults to None.
        ylabels: labels for the rows (header). Defaults to None.
        variable_size: Flag to indicate if the array dimensions can be modified.
         If a BaseArrayHandler is given as input, the handler should also be in
         readonly mode Defaults to False.
        current_slice: slice of the same dimension as the Numpy ndarray that will.
         Defaults to None
    """

    # _data: MaskedArrayHandler

    def __init__(
        self,
        parent: QWidget,
        data: np.ma.MaskedArray | MaskedArrayHandler,
        readonly=False,
        xlabels=None,
        ylabels=None,
        variable_size=False,
        current_slice: Sequence[slice | int] | None = None,
    ) -> None:
        super().__init__(
            parent, data, readonly, xlabels, ylabels, variable_size, current_slice
        )

    def _init_model(
        self,
        xlabels: Sequence[str] | None,
        ylabels: Sequence[str] | None,
        readonly: bool,
        current_slice: Sequence[slice | int] | None = None,
    ) -> None:
        """Initializes and set the instance model to use

        Args:
            xlabels: labels for the columns (header). Defaults to None.
            ylabels: labels for the rows (header). Defaults to None.
            readonly: Flag for readonly mode. Defaults to False.
            current_slice: slice of the same dimension as the Numpy ndarray that will.
        """
        assert isinstance(self._data, MaskedArrayHandler)
        self.model = DataArrayModel(
            self._data,
            xlabels=xlabels,
            ylabels=ylabels,
            readonly=readonly,
            parent=self,
            current_slice=current_slice,
        )


class RecordArrayEditorWidget(BaseArrayEditorWidget):
    """Same as BaseArrayWidgetEditorWidget but specifically handles RecordArrayHandler
    and RecordArrayModel which are made to wrap Numpy's structured arrays.

    Args:
        parent: parent QObject
        data: Numpy's ndarray or BaseArrayHandler to use.
        readonly: Flag for readonly mode. Defaults to False.
        xlabels: labels for the columns (header). Defaults to None.
        ylabels: labels for the rows (header). Defaults to None.
        variable_size: Flag to indicate if the array dimensions can be modified.
         If a BaseArrayHandler is given as input, the handler should also be in
         readonly mode Defaults to False.
        current_slice: slice of the same dimension as the Numpy ndarray that will.
         Defaults to None
    """

    def __init__(
        self,
        parent,
        data: RecordArrayHandler,
        dtype_name: str,
        readonly=False,
        xlabels=None,
        ylabels=None,
        variable_size=False,
        current_slice: Sequence[slice | int] | None = None,
    ) -> None:
        self._dtype_name = dtype_name
        super().__init__(
            parent, data, readonly, xlabels, ylabels, variable_size, current_slice
        )

    def _init_handler(self, data: np.ndarray | RecordArrayHandler) -> None:
        """Initializes and set the instance handler to use

        Args:
            data: Numpy's ndarray or BaseArrayHandler to use.
        """
        if isinstance(data, np.ma.MaskedArray):
            self._data = RecordArrayHandler(data, self._variable_size)
        elif isinstance(data, RecordArrayHandler):
            self._data = data
        else:
            raise TypeError(
                f"Given data must be of type np.ndarray or RecordArrayHandler, not {type(data)}"
            )

    def _init_model(
        self,
        xlabels: Sequence[str] | None,
        ylabels: Sequence[str] | None,
        readonly: bool,
        current_slice: Sequence[slice | int] | None = None,
    ) -> None:
        """Initializes and set the instance model to use

        Args:
            xlabels: labels for the columns (header). Defaults to None.
            ylabels: labels for the rows (header). Defaults to None.
            readonly: Flag for readonly mode. Defaults to False.
            current_slice: slice of the same dimension as the Numpy ndarray that will.
        """
        assert isinstance(self._data, RecordArrayHandler)
        self.model = RecordArrayModel(
            self._data,
            self._dtype_name,
            xlabels=xlabels,
            ylabels=ylabels,
            readonly=readonly,
            parent=self,
            current_slice=current_slice,
        )



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Same as test_all_items.py but with readonly=True to check that the readonly mode works
on all DataItem types.
"""

# guitest: show

import numpy as np

from guidata.env import execenv
from guidata.qthelpers import qt_app_context
from guidata.tests.dataset.test_all_items import Parameters

# check if array variable_size is ignored thanks to readonly
Parameters.floatarray.set_prop("edit", variable_size=True)


def test_all_features():
    """Test all guidata item/group features"""
    with qt_app_context():
        prm1 = Parameters(readonly=True)
        prm1.floatarray[:, 0] = np.linspace(-5, 5, 50)
        execenv.print(prm1)

        if prm1.edit():
            prm1.edit()
            execenv.print(prm1)
        prm1.view()

        prm2 = Parameters.create(integer=10101010, string="Using create class method")
        print(prm2)

        execenv.print("OK")


if __name__ == "__main__":
    test_all_features()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2010 IPython Development Team
# Copyright (c) 2013- Spyder Project Contributors
#
# Distributed under the terms of the Modified BSD License
# (BSD 3-clause; see NOTICE.txt in the Spyder root directory for details).
# -----------------------------------------------------------------------------

"""
Calltip widget used only to show signatures.

Adapted from IPython/frontend/qt/console/call_tip_widget.py of the
`IPython Project <https://github.com/ipython/ipython>`_.
Now located at qtconsole/call_tip_widget.py as part of the
`Jupyter QtConsole Project <https://github.com/jupyter/qtconsole>`_.
"""

from unicodedata import category

from qtpy.QtCore import QBasicTimer, QCoreApplication, QEvent, Qt
from qtpy.QtGui import QCursor, QPalette
from qtpy.QtWidgets import (
    QFrame,
    QLabel,
    QPlainTextEdit,
    QStyle,
    QStyleOptionFrame,
    QStylePainter,
    QTextEdit,
    QToolTip,
)


class CallTipWidget(QLabel):
    """Shows call tips by parsing the current text of Q[Plain]TextEdit."""

    # --------------------------------------------------------------------------
    # 'QObject' interface
    # --------------------------------------------------------------------------

    def __init__(self, text_edit, hide_timer_on=False):
        """Create a call tip manager that is attached to the specified Qt
        text edit widget.
        """
        assert isinstance(text_edit, (QTextEdit, QPlainTextEdit))
        super().__init__(None, Qt.ToolTip)
        self.app = QCoreApplication.instance()

        self.hide_timer_on = hide_timer_on
        self.tip = None
        self._hide_timer = QBasicTimer()
        self._text_edit = text_edit

        self.setFont(text_edit.document().defaultFont())
        self.setForegroundRole(QPalette.ToolTipText)
        self.setBackgroundRole(QPalette.ToolTipBase)
        self.setPalette(QToolTip.palette())

        self.setAlignment(Qt.AlignLeft)
        self.setIndent(1)
        self.setFrameStyle(QFrame.NoFrame)
        self.setMargin(
            1 + self.style().pixelMetric(QStyle.PM_ToolTipLabelFrameWidth, None, self)
        )

    def eventFilter(self, obj, event):
        """Reimplemented to hide on certain key presses and on text edit focus
        changes.
        """
        if obj == self._text_edit:
            etype = event.type()

            if etype == QEvent.KeyPress:
                key = event.key()
                cursor = self._text_edit.textCursor()
                prev_char = self._text_edit.get_character(cursor.position(), offset=-1)
                if key in (Qt.Key_Enter, Qt.Key_Return, Qt.Key_Down, Qt.Key_Up):
                    self.hide()
                elif key == Qt.Key_Escape:
                    self.hide()
                    return True
                elif prev_char == ")":
                    self.hide()

            elif etype == QEvent.FocusOut:
                self.hide()

            elif etype == QEvent.Enter:
                if (
                    self._hide_timer.isActive()
                    and self.app.topLevelAt(QCursor.pos()) == self
                ):
                    self._hide_timer.stop()

            elif etype == QEvent.Leave:
                self._leave_event_hide()

        return super().eventFilter(obj, event)

    def timerEvent(self, event):
        """Reimplemented to hide the widget when the hide timer fires."""
        if event.timerId() == self._hide_timer.timerId():
            self._hide_timer.stop()
            self.hide()

    # --------------------------------------------------------------------------
    # 'QWidget' interface
    # --------------------------------------------------------------------------

    def enterEvent(self, event):
        """Reimplemented to cancel the hide timer."""
        super().enterEvent(event)
        if self._hide_timer.isActive() and self.app.topLevelAt(QCursor.pos()) == self:
            self._hide_timer.stop()

    def hideEvent(self, event):
        """Reimplemented to disconnect signal handlers and event filter."""
        super().hideEvent(event)
        self._text_edit.cursorPositionChanged.disconnect(self._cursor_position_changed)
        self._text_edit.removeEventFilter(self)

    def leaveEvent(self, event):
        """Reimplemented to start the hide timer."""
        super().leaveEvent(event)
        self._leave_event_hide()

    def mousePressEvent(self, event):
        """

        :param event:
        """
        super().mousePressEvent(event)
        self.hide()

    def paintEvent(self, event):
        """Reimplemented to paint the background panel."""
        painter = QStylePainter(self)
        option = QStyleOptionFrame()
        option.initFrom(self)
        painter.drawPrimitive(QStyle.PE_PanelTipLabel, option)
        painter.end()

        super().paintEvent(event)

    def setFont(self, font):
        """Reimplemented to allow use of this method as a slot."""
        super().setFont(font)

    def showEvent(self, event):
        """Reimplemented to connect signal handlers and event filter."""
        super().showEvent(event)
        self._text_edit.cursorPositionChanged.connect(self._cursor_position_changed)
        self._text_edit.installEventFilter(self)

    def focusOutEvent(self, event):
        """Reimplemented to hide it when focus goes out of the main
        window.
        """
        self.hide()

    # --------------------------------------------------------------------------
    # 'CallTipWidget' interface
    # --------------------------------------------------------------------------

    def show_tip(self, point, tip, wrapped_tiplines):
        """Attempts to show the specified tip at the current cursor location."""
        # Don't attempt to show it if it's already visible and the text
        # to be displayed is the same as the one displayed before.
        if self.isVisible():
            if self.tip == tip:
                return True
            else:
                self.hide()

        # Attempt to find the cursor position at which to show the call tip.
        text_edit = self._text_edit
        cursor = text_edit.textCursor()
        search_pos = cursor.position() - 1
        self._start_position, _ = self._find_parenthesis(search_pos, forward=False)
        if self._start_position == -1:
            return False

        if self.hide_timer_on:
            self._hide_timer.stop()
            # Logic to decide how much time to show the calltip depending
            # on the amount of text present
            if len(wrapped_tiplines) == 1:
                args = wrapped_tiplines[0].split("(")[1]
                nargs = len(args.split(","))
                if nargs == 1:
                    hide_time = 1400
                elif nargs == 2:
                    hide_time = 1600
                else:
                    hide_time = 1800
            elif len(wrapped_tiplines) == 2:
                args1 = wrapped_tiplines[1].strip()
                nargs1 = len(args1.split(","))
                if nargs1 == 1:
                    hide_time = 2500
                else:
                    hide_time = 2800
            else:
                hide_time = 3500
            self._hide_timer.start(hide_time, self)

        # Set the text and resize the widget accordingly.
        self.tip = tip
        self.setText(tip)
        self.resize(self.sizeHint())

        # Locate and show the widget. Place the tip below the current line
        # unless it would be off the screen.  In that case, decide the best
        # location based trying to minimize the  area that goes off-screen.
        padding = 3  # Distance in pixels between cursor bounds and tip box.
        cursor_rect = text_edit.cursorRect(cursor)
        screen_rect = text_edit.screen().geometry()
        point.setY(point.y() + padding)
        tip_height = self.size().height()
        tip_width = self.size().width()

        vertical = "bottom"
        horizontal = "Right"
        if point.y() + tip_height > screen_rect.height() + screen_rect.y():
            point_ = text_edit.mapToGlobal(cursor_rect.topRight())
            # If tip is still off screen, check if point is in top or bottom
            # half of screen.
            if point_.y() - tip_height < padding:
                # If point is in upper half of screen, show tip below it.
                # otherwise above it.
                if 2 * point.y() < screen_rect.height():
                    vertical = "bottom"
                else:
                    vertical = "top"
            else:
                vertical = "top"
        if point.x() + tip_width > screen_rect.width() + screen_rect.x():
            point_ = text_edit.mapToGlobal(cursor_rect.topRight())
            # If tip is still off-screen, check if point is in the right or
            # left half of the screen.
            if point_.x() - tip_width < padding:
                if 2 * point.x() < screen_rect.width():
                    horizontal = "Right"
                else:
                    horizontal = "Left"
            else:
                horizontal = "Left"
        pos = getattr(cursor_rect, "%s%s" % (vertical, horizontal))
        adjusted_point = text_edit.mapToGlobal(pos())
        if vertical == "top":
            point.setY(adjusted_point.y() - tip_height - padding)
        if horizontal == "Left":
            point.setX(adjusted_point.x() - tip_width - padding)

        self.move(point)
        self.show()
        return True

    # --------------------------------------------------------------------------
    # Protected interface
    # --------------------------------------------------------------------------

    def _find_parenthesis(self, position, forward=True):
        """If 'forward' is True (resp. False), proceed forwards
        (resp. backwards) through the line that contains 'position' until an
        unmatched closing (resp. opening) parenthesis is found. Returns a
        tuple containing the position of this parenthesis (or -1 if it is
        not found) and the number commas (at depth 0) found along the way.
        """
        commas = depth = 0
        document = self._text_edit.document()
        char = str(document.characterAt(position))
        # Search until a match is found or a non-printable character is
        # encountered.
        while category(char) != "Cc" and position > 0:
            if char == "," and depth == 0:
                commas += 1
            elif char == ")":
                if forward and depth == 0:
                    break
                depth += 1
            elif char == "(":
                if not forward and depth == 0:
                    break
                depth -= 1
            position += 1 if forward else -1
            char = str(document.characterAt(position))
        else:
            position = -1
        return position, commas

    def _leave_event_hide(self):
        """Hides the tooltip after some time has passed (assuming the cursor is
        not over the tooltip).
        """
        if (
            self.hide_timer_on
            and not self._hide_timer.isActive()
            and
            # If Enter events always came after Leave events, we wouldn't need
            # this check. But on Mac OS, it sometimes happens the other way
            # around when the tooltip is created.
            self.app.topLevelAt(QCursor.pos()) != self
        ):
            self._hide_timer.start(800, self)

    # ------ Signal handlers ----------------------------------------------------

    def _cursor_position_changed(self):
        """Updates the tip based on user cursor movement."""
        cursor = self._text_edit.textCursor()
        position = cursor.position()
        document = self._text_edit.document()
        char = str(document.characterAt(position - 1))
        if position <= self._start_position:
            self.hide()
        elif char == ")":
            pos, _ = self._find_parenthesis(position - 1, forward=False)
            if pos == -1:
                self.hide()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""Config test"""

import pytest

from guidata.config import UserConfig
from guidata.tests.dataset.test_all_features import Parameters


@pytest.fixture()
def config():
    """Create a config object"""
    CONF = UserConfig({})
    eta = Parameters()
    eta.write_config(CONF, "TestParameters", "")
    yield CONF


def test_load(config):
    """Test load config"""
    eta = Parameters()
    eta.read_config(config, "TestParameters", "")


def test_default(config):
    """Test default config"""
    eta = Parameters()
    eta.write_config(config, "etagere2", "")
    eta = Parameters()
    eta.read_config(config, "etagere2", "")


def test_restore(config):
    """Test restore config"""
    eta = Parameters()
    eta.fl2 = 2
    eta.integer = 6
    eta.write_config(config, "etagere3", "")

    eta = Parameters()
    eta.read_config(config, "etagere3", "")

    assert eta.fl2 == 2.0
    assert eta.integer == 6



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
DataSet item order test

From time to time, it may be useful to change the item order,
for example when deriving a dataset from another.
"""

# guitest: show

import guidata.dataset as gds
from guidata.env import execenv
from guidata.qthelpers import qt_app_context


class OriginalDataset(gds.DataSet):
    """Original dataset
    This is the original dataset"""

    param1 = gds.BoolItem("P1 | Boolean")
    param2 = gds.StringItem("P2 | String")
    param3 = gds.TextItem("P3 | Text")
    param4 = gds.FloatItem("P4 | Float", default=0)


class DerivedDataset(OriginalDataset):
    """Derived dataset
    This is the derived dataset, with modified item order"""

    param5 = gds.IntItem("P5 | Int", default=0).set_pos(row=2)
    param6 = gds.DateItem("P6 | Date", default=0).set_pos(row=4)


def test_item_order():
    """Test DataSet item order"""
    with qt_app_context():
        e = OriginalDataset()
        e.edit()
        execenv.print(e)

        e = DerivedDataset()
        e.edit()
        execenv.print(e)
        execenv.print("OK")


class BaseDataset(gds.DataSet):
    """Base dataset for inheritance ordering test"""

    base_param1 = gds.StringItem("Base Param 1", default="base1")
    base_param2 = gds.StringItem("Base Param 2", default="base2")


class MiddleDataset(BaseDataset):
    """Middle dataset that overrides a base parameter"""

    base_param1 = gds.StringItem("Base Param 1", default="overridden")  # Override
    middle_param = gds.FloatItem("Middle Param", default=1.0)


class ChildDataset(MiddleDataset):
    """Child dataset that adds more parameters"""

    child_param1 = gds.IntItem("Child Param 1", default=42)
    child_param2 = gds.BoolItem("Child Param 2", default=True)


def test_inheritance_item_ordering_and_overrides():
    """Test that DataSet inheritance preserves correct item ordering and overrides

    This test ensures that:
    1. Parent class items appear before child class items in the final dataset
    2. Attribute redefinition in intermediate classes properly overrides parent
       values
    3. Child classes inherit the overridden values, not the original parent
       values

    This prevents regression of the inheritance bug where child classes would get
    the original grandparent values instead of the redefined parent values.
    """
    # Test the child dataset
    child = ChildDataset()

    # Get the item names in order
    item_names = [item._name for item in child._items]

    # Verify ordering: BaseDataset items first, then MiddleDataset items,
    # then ChildDataset items
    expected_order = [
        "base_param1",  # From BaseDataset (but overridden by MiddleDataset)
        "base_param2",  # From BaseDataset
        "middle_param",  # From MiddleDataset
        "child_param1",  # From ChildDataset
        "child_param2",  # From ChildDataset
    ]

    # Assert item ordering is correct
    assert item_names == expected_order, (
        f"Item ordering is incorrect. Expected: {expected_order}, Got: {item_names}"
    )

    # Verify that override values are correctly inherited
    assert child.base_param1 == "overridden", (
        f"Override inheritance failed. Expected 'overridden', got '{child.base_param1}'"
    )
    assert child.base_param2 == "base2", (
        f"Normal inheritance failed. Expected 'base2', got '{child.base_param2}'"
    )
    assert child.middle_param == 1.0, (
        f"Middle class value failed. Expected 1.0, got {child.middle_param}"
    )

    # Test that direct middle class also has correct values
    middle = MiddleDataset()
    assert middle.base_param1 == "overridden", (
        f"Middle class override failed. Expected 'overridden', "
        f"got '{middle.base_param1}'"
    )
    assert middle.base_param2 == "base2", (
        f"Middle class inheritance failed. Expected 'base2', got '{middle.base_param2}'"
    )

    # Test that base class has original values
    base = BaseDataset()
    assert base.base_param1 == "base1", (
        f"Base class value changed. Expected 'base1', got '{base.base_param1}'"
    )
    assert base.base_param2 == "base2", (
        f"Base class value changed. Expected 'base2', got '{base.base_param2}'"
    )

    execenv.print("Inheritance item ordering and overrides test: OK")


if __name__ == "__main__":
    test_inheritance_item_ordering_and_overrides()
    test_item_order()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# ruff: noqa

"""
Text data Importing Wizard based on Qt
"""

# ----date and datetime objects support
import datetime
import io
from functools import partial as ft_partial
from itertools import zip_longest

from numpy import nan
from qtpy.QtCore import QAbstractTableModel, QModelIndex, Qt, Signal, Slot
from qtpy.QtGui import QColor, QIntValidator
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QFrame,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMenu,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QSizePolicy,
    QSpacerItem,
    QTableView,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from guidata.config import _
from guidata.configtools import get_icon
from guidata.qthelpers import add_actions, create_action, win32_fix_title_bar_background

try:
    import pandas as pd
except:
    pd = None


def try_to_parse(value):
    """

    :param value:
    :return:
    """
    _types = ("int", "float")
    for _t in _types:
        try:
            _val = eval("%s('%s')" % (_t, value))
            return _val
        except (ValueError, SyntaxError):
            pass
    return value


def try_to_eval(value):
    """

    :param value:
    :return:
    """
    try:
        return eval(value)
    except (NameError, SyntaxError, ImportError):
        return value


# ----Numpy arrays support
class FakeObject:
    """Fake class used in replacement of missing modules"""

    pass


try:
    from numpy import array, ndarray
except:

    class ndarray(FakeObject):  # analysis:ignore
        """Fake ndarray"""

        pass


try:
    from dateutil.parser import parse as dateparse
except:

    def dateparse(datestr, dayfirst=True):  # analysis:ignore
        """Just for 'day/month/year' strings"""
        _a, _b, _c = list(map(int, datestr.split("/")))
        if dayfirst:
            return datetime.datetime(_c, _b, _a)
        return datetime.datetime(_c, _a, _b)


def datestr_to_datetime(value, dayfirst=True):
    """

    :param value:
    :param dayfirst:
    :return:
    """
    return dateparse(value, dayfirst=dayfirst)


# ----Background colors for supported types
COLORS = {
    bool: Qt.magenta,
    (float, int): Qt.blue,
    list: Qt.yellow,
    dict: Qt.cyan,
    tuple: Qt.lightGray,
    (str,): Qt.darkRed,
    ndarray: Qt.green,
    datetime.date: Qt.darkYellow,
}


def get_color(value, alpha):
    """Return color depending on value type"""
    color = QColor()
    for typ in COLORS:
        if isinstance(value, typ):
            color = QColor(COLORS[typ])
    color.setAlphaF(alpha)
    return color


class ContentsWidget(QWidget):
    """Import wizard contents widget"""

    asDataChanged = Signal(bool)

    def __init__(self, parent, text):
        QWidget.__init__(self, parent)

        self.text_editor = QTextEdit(self)
        self.text_editor.setText(text)
        self.text_editor.setReadOnly(True)

        # Type frame
        type_layout = QHBoxLayout()
        type_label = QLabel(_("Import as"))
        type_layout.addWidget(type_label)
        data_btn = QRadioButton(_("data"))
        data_btn.setChecked(True)
        self._as_data = True
        type_layout.addWidget(data_btn)
        code_btn = QRadioButton(_("code"))
        self._as_code = False
        type_layout.addWidget(code_btn)
        txt_btn = QRadioButton(_("text"))
        type_layout.addWidget(txt_btn)

        h_spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        type_layout.addItem(h_spacer)
        type_frame = QFrame()
        type_frame.setLayout(type_layout)

        # Opts frame
        grid_layout = QGridLayout()
        grid_layout.setSpacing(0)

        col_label = QLabel(_("Column separator:"))
        grid_layout.addWidget(col_label, 0, 0)
        col_w = QWidget()
        col_btn_layout = QHBoxLayout()
        self.tab_btn = QRadioButton(_("Tab"))
        self.tab_btn.setChecked(False)
        col_btn_layout.addWidget(self.tab_btn)
        self.ws_btn = QRadioButton(_("Whitespace"))
        self.ws_btn.setChecked(False)
        col_btn_layout.addWidget(self.ws_btn)
        other_btn_col = QRadioButton(_("other"))
        other_btn_col.setChecked(True)
        col_btn_layout.addWidget(other_btn_col)
        col_w.setLayout(col_btn_layout)
        grid_layout.addWidget(col_w, 0, 1)
        self.line_edt = QLineEdit(",")
        self.line_edt.setMaximumWidth(30)
        self.line_edt.setEnabled(True)
        other_btn_col.toggled.connect(self.line_edt.setEnabled)
        grid_layout.addWidget(self.line_edt, 0, 2)

        row_label = QLabel(_("Row separator:"))
        grid_layout.addWidget(row_label, 1, 0)
        row_w = QWidget()
        row_btn_layout = QHBoxLayout()
        self.eol_btn = QRadioButton(_("EOL"))
        self.eol_btn.setChecked(True)
        row_btn_layout.addWidget(self.eol_btn)
        other_btn_row = QRadioButton(_("other"))
        row_btn_layout.addWidget(other_btn_row)
        row_w.setLayout(row_btn_layout)
        grid_layout.addWidget(row_w, 1, 1)
        self.line_edt_row = QLineEdit(";")
        self.line_edt_row.setMaximumWidth(30)
        self.line_edt_row.setEnabled(False)
        other_btn_row.toggled.connect(self.line_edt_row.setEnabled)
        grid_layout.addWidget(self.line_edt_row, 1, 2)

        grid_layout.setRowMinimumHeight(2, 15)

        other_group = QGroupBox(_("Additional options"))
        other_layout = QGridLayout()
        other_group.setLayout(other_layout)

        skiprows_label = QLabel(_("Skip rows:"))
        other_layout.addWidget(skiprows_label, 0, 0)
        self.skiprows_edt = QLineEdit("0")
        self.skiprows_edt.setMaximumWidth(30)
        intvalid = QIntValidator(0, len(str(text).splitlines()), self.skiprows_edt)
        self.skiprows_edt.setValidator(intvalid)
        other_layout.addWidget(self.skiprows_edt, 0, 1)

        other_layout.setColumnMinimumWidth(2, 5)

        comments_label = QLabel(_("Comments:"))
        other_layout.addWidget(comments_label, 0, 3)
        self.comments_edt = QLineEdit("#")
        self.comments_edt.setMaximumWidth(30)
        other_layout.addWidget(self.comments_edt, 0, 4)

        self.trnsp_box = QCheckBox(_("Transpose"))
        # self.trnsp_box.setEnabled(False)
        other_layout.addWidget(self.trnsp_box, 1, 0, 2, 0)

        grid_layout.addWidget(other_group, 3, 0, 2, 0)

        opts_frame = QFrame()
        opts_frame.setLayout(grid_layout)

        data_btn.toggled.connect(opts_frame.setEnabled)
        data_btn.toggled.connect(self.set_as_data)
        code_btn.toggled.connect(self.set_as_code)
        #        self.connect(txt_btn, SIGNAL("toggled(bool)"),
        #                     self, SLOT("is_text(bool)"))

        # Final layout
        layout = QVBoxLayout()
        layout.addWidget(type_frame)
        layout.addWidget(self.text_editor)
        layout.addWidget(opts_frame)
        self.setLayout(layout)

    def get_as_data(self):
        """Return if data type conversion"""
        return self._as_data

    def get_as_code(self):
        """Return if code type conversion"""
        return self._as_code

    def get_as_num(self):
        """Return if numeric type conversion"""
        return self._as_num

    def get_col_sep(self):
        """Return the column separator"""
        if self.tab_btn.isChecked():
            return "\t"
        elif self.ws_btn.isChecked():
            return None
        return str(self.line_edt.text())

    def get_row_sep(self):
        """Return the row separator"""
        if self.eol_btn.isChecked():
            return "\n"
        return str(self.line_edt_row.text())

    def get_skiprows(self):
        """Return number of lines to be skipped"""
        return int(str(self.skiprows_edt.text()))

    def get_comments(self):
        """Return comment string"""
        return str(self.comments_edt.text())

    @Slot(bool)
    def set_as_data(self, as_data):
        """Set if data type conversion"""
        self._as_data = as_data
        self.asDataChanged.emit(as_data)

    @Slot(bool)
    def set_as_code(self, as_code):
        """Set if code type conversion"""
        self._as_code = as_code


class PreviewTableModel(QAbstractTableModel):
    """Import wizard preview table model"""

    def __init__(self, data=[], parent=None):
        QAbstractTableModel.__init__(self, parent)
        self._data = data

    def rowCount(self, parent=QModelIndex()):
        """Return row count"""
        return len(self._data)

    def columnCount(self, parent=QModelIndex()):
        """Return column count"""
        return len(self._data[0])

    def _display_data(self, index):
        """Return a data element"""
        return self._data[index.row()][index.column()]

    def data(self, index, role=Qt.DisplayRole):
        """Return a model data element"""
        if not index.isValid():
            return None
        if role == Qt.DisplayRole:
            return self._display_data(index)
        elif role == Qt.BackgroundColorRole:
            return get_color(self._data[index.row()][index.column()], 0.2)
        elif role == Qt.TextAlignmentRole:
            return int(Qt.AlignRight | Qt.AlignVCenter)
        return None

    def setData(self, index, value, role=Qt.EditRole):
        """Set model data"""
        return False

    def get_data(self):
        """Return a copy of model data"""
        return self._data[:][:]

    def parse_data_type(self, index, **kwargs):
        """Parse a type to an other type"""
        if not index.isValid():
            return False
        try:
            if kwargs["atype"] == "date":
                self._data[index.row()][index.column()] = datestr_to_datetime(
                    self._data[index.row()][index.column()], kwargs["dayfirst"]
                ).date()
            elif kwargs["atype"] == "perc":
                _tmp = self._data[index.row()][index.column()].replace("%", "")
                self._data[index.row()][index.column()] = eval(_tmp) / 100.0
            elif kwargs["atype"] == "account":
                _tmp = self._data[index.row()][index.column()].replace(",", "")
                self._data[index.row()][index.column()] = eval(_tmp)
            elif kwargs["atype"] == "unicode":
                self._data[index.row()][index.column()] = str(
                    self._data[index.row()][index.column()]
                )
            elif kwargs["atype"] == "int":
                self._data[index.row()][index.column()] = int(
                    self._data[index.row()][index.column()]
                )
            elif kwargs["atype"] == "float":
                self._data[index.row()][index.column()] = float(
                    self._data[index.row()][index.column()]
                )
            self.dataChanged.emit(index, index)
        except Exception as instance:
            print(instance)  # spyder: test-skip

    def reset(self):
        """ """
        self.beginResetModel()
        self.endResetModel()


class PreviewTable(QTableView):
    """Import wizard preview widget"""

    def __init__(self, parent):
        QTableView.__init__(self, parent)
        self._model = None

        # Setting up actions
        self.date_dayfirst_action = create_action(
            self,
            "dayfirst",
            triggered=ft_partial(self.parse_to_type, atype="date", dayfirst=True),
        )
        self.date_monthfirst_action = create_action(
            self,
            "monthfirst",
            triggered=ft_partial(self.parse_to_type, atype="date", dayfirst=False),
        )
        self.perc_action = create_action(
            self, "perc", triggered=ft_partial(self.parse_to_type, atype="perc")
        )
        self.acc_action = create_action(
            self, "account", triggered=ft_partial(self.parse_to_type, atype="account")
        )
        self.str_action = create_action(
            self, "unicode", triggered=ft_partial(self.parse_to_type, atype="unicode")
        )
        self.int_action = create_action(
            self, "int", triggered=ft_partial(self.parse_to_type, atype="int")
        )
        self.float_action = create_action(
            self, "float", triggered=ft_partial(self.parse_to_type, atype="float")
        )

        # Setting up menus
        self.date_menu = QMenu()
        self.date_menu.setTitle("Date")
        add_actions(
            self.date_menu, (self.date_dayfirst_action, self.date_monthfirst_action)
        )
        self.parse_menu = QMenu(self)
        self.parse_menu.addMenu(self.date_menu)
        add_actions(self.parse_menu, (self.perc_action, self.acc_action))
        self.parse_menu.setTitle("String to")
        self.opt_menu = QMenu(self)
        self.opt_menu.addMenu(self.parse_menu)
        add_actions(
            self.opt_menu, (self.str_action, self.int_action, self.float_action)
        )

    def _shape_text(
        self,
        text,
        colsep="\t",
        rowsep="\n",
        transpose=False,
        skiprows=0,
        comments="#",
    ):
        """Decode the shape of the given text"""
        assert colsep != rowsep
        out = []
        text_rows = text.split(rowsep)[skiprows:]
        for row in text_rows:
            stripped = str(row).strip()
            if len(stripped) == 0 or stripped.startswith(comments):
                continue
            line = str(row).split(colsep)
            line = [try_to_parse(str(x)) for x in line]
            out.append(line)
        # Replace missing elements with np.nan's or None's

        out = list(zip_longest(*out, fillvalue=nan))

        # Tranpose the last result to get the expected one
        out = [[r[col] for r in out] for col in range(len(out[0]))]
        if transpose:
            return [[r[col] for r in out] for col in range(len(out[0]))]
        return out

    def get_data(self):
        """Return model data"""
        if self._model is None:
            return None
        return self._model.get_data()

    def process_data(
        self,
        text,
        colsep="\t",
        rowsep="\n",
        transpose=False,
        skiprows=0,
        comments="#",
    ):
        """Put data into table model"""
        data = self._shape_text(text, colsep, rowsep, transpose, skiprows, comments)
        self._model = PreviewTableModel(data)
        self.setModel(self._model)

    @Slot()
    def parse_to_type(self, **kwargs):
        """Parse to a given type"""
        indexes = self.selectedIndexes()
        if not indexes:
            return
        for index in indexes:
            self.model().parse_data_type(index, **kwargs)

    def contextMenuEvent(self, event):
        """Reimplement Qt method"""
        self.opt_menu.popup(event.globalPos())
        event.accept()


class PreviewWidget(QWidget):
    """Import wizard preview widget"""

    def __init__(self, parent):
        QWidget.__init__(self, parent)

        vert_layout = QVBoxLayout()

        # Type frame
        type_layout = QHBoxLayout()
        type_label = QLabel(_("Import as"))
        type_layout.addWidget(type_label)

        self.array_btn = array_btn = QRadioButton(_("array"))
        array_btn.setEnabled(ndarray is not FakeObject)
        array_btn.setChecked(ndarray is not FakeObject)
        type_layout.addWidget(array_btn)

        list_btn = QRadioButton(_("list"))
        list_btn.setChecked(not array_btn.isChecked())
        type_layout.addWidget(list_btn)

        if pd:
            self.df_btn = df_btn = QRadioButton(_("DataFrame"))
            df_btn.setChecked(False)
            type_layout.addWidget(df_btn)

        h_spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        type_layout.addItem(h_spacer)
        type_frame = QFrame()
        type_frame.setLayout(type_layout)

        self._table_view = PreviewTable(self)
        vert_layout.addWidget(type_frame)
        vert_layout.addWidget(self._table_view)
        self.setLayout(vert_layout)

    def open_data(
        self,
        text,
        colsep="\t",
        rowsep="\n",
        transpose=False,
        skiprows=0,
        comments="#",
    ):
        """Open clipboard text as table"""
        if pd:
            self.pd_text = text
            self.pd_info = dict(
                sep=colsep, lineterminator=rowsep, skiprows=skiprows, comment=comments
            )
            if colsep is None:
                self.pd_info = dict(
                    lineterminator=rowsep,
                    skiprows=skiprows,
                    comment=comments,
                    delim_whitespace=True,
                )
        self._table_view.process_data(
            text, colsep, rowsep, transpose, skiprows, comments
        )

    def get_data(self):
        """Return table data"""
        return self._table_view.get_data()


class ImportWizard(QDialog):
    """Text data import wizard"""

    def __init__(
        self, parent, text, title=None, icon=None, contents_title=None, varname=None
    ):
        QDialog.__init__(self, parent)

        win32_fix_title_bar_background(self)

        # Destroying the C++ object right after closing the dialog box,
        # otherwise it may be garbage-collected in another QThread
        # (e.g. the editor's analysis thread in Spyder), thus leading to
        # a segmentation fault on UNIX or an application crash on Windows
        self.setAttribute(Qt.WA_DeleteOnClose)

        if title is None:
            title = _("Import wizard")
        self.setWindowTitle(title)
        if icon is None:
            self.setWindowIcon(get_icon("fileimport.png"))
        if contents_title is None:
            contents_title = _("Raw text")

        if varname is None:
            varname = _("variable_name")

        self.var_name, self.clip_data = None, None

        # Setting GUI
        self.tab_widget = QTabWidget(self)
        self.text_widget = ContentsWidget(self, text)
        self.table_widget = PreviewWidget(self)

        self.tab_widget.addTab(self.text_widget, _("text"))
        self.tab_widget.setTabText(0, contents_title)
        self.tab_widget.addTab(self.table_widget, _("table"))
        self.tab_widget.setTabText(1, _("Preview"))
        self.tab_widget.setTabEnabled(1, False)

        name_layout = QHBoxLayout()
        name_label = QLabel(_("Variable Name"))
        name_layout.addWidget(name_label)

        self.name_edt = QLineEdit()
        self.name_edt.setText(varname)
        name_layout.addWidget(self.name_edt)

        btns_layout = QHBoxLayout()
        cancel_btn = QPushButton(_("Cancel"))
        btns_layout.addWidget(cancel_btn)
        cancel_btn.clicked.connect(self.reject)
        h_spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        btns_layout.addItem(h_spacer)
        self.back_btn = QPushButton(_("Previous"))
        self.back_btn.setEnabled(False)
        btns_layout.addWidget(self.back_btn)
        self.back_btn.clicked.connect(ft_partial(self._set_step, step=-1))
        self.fwd_btn = QPushButton(_("Next"))
        if not text:
            self.fwd_btn.setEnabled(False)
        btns_layout.addWidget(self.fwd_btn)
        self.fwd_btn.clicked.connect(ft_partial(self._set_step, step=1))
        self.done_btn = QPushButton(_("Done"))
        self.done_btn.setEnabled(False)
        btns_layout.addWidget(self.done_btn)
        self.done_btn.clicked.connect(self.process)

        self.text_widget.asDataChanged.connect(self.fwd_btn.setEnabled)
        self.text_widget.asDataChanged.connect(self.done_btn.setDisabled)
        layout = QVBoxLayout()
        layout.addLayout(name_layout)
        layout.addWidget(self.tab_widget)
        layout.addLayout(btns_layout)
        self.setLayout(layout)

    def _focus_tab(self, tab_idx):
        """Change tab focus"""
        for i in range(self.tab_widget.count()):
            self.tab_widget.setTabEnabled(i, False)
        self.tab_widget.setTabEnabled(tab_idx, True)
        self.tab_widget.setCurrentIndex(tab_idx)

    def _set_step(self, step):
        """Proceed to a given step"""
        new_tab = self.tab_widget.currentIndex() + step
        assert new_tab < self.tab_widget.count() and new_tab >= 0
        if new_tab == self.tab_widget.count() - 1:
            try:
                self.table_widget.open_data(
                    self._get_plain_text(),
                    self.text_widget.get_col_sep(),
                    self.text_widget.get_row_sep(),
                    self.text_widget.trnsp_box.isChecked(),
                    self.text_widget.get_skiprows(),
                    self.text_widget.get_comments(),
                )
                self.done_btn.setEnabled(True)
                self.done_btn.setDefault(True)
                self.fwd_btn.setEnabled(False)
                self.back_btn.setEnabled(True)
            except (SyntaxError, AssertionError) as error:
                QMessageBox.critical(
                    self,
                    _("Import wizard"),
                    _(
                        "<b>Unable to proceed to next step</b>"
                        "<br><br>Please check your entries."
                        "<br><br>Error message:<br>%s"
                    )
                    % str(error),
                )
                return
        elif new_tab == 0:
            self.done_btn.setEnabled(False)
            self.fwd_btn.setEnabled(True)
            self.back_btn.setEnabled(False)
        self._focus_tab(new_tab)

    def get_data(self):
        """Return processed data"""
        # It is import to avoid accessing Qt C++ object as it has probably
        # already been destroyed, due to the Qt.WA_DeleteOnClose attribute
        return self.var_name, self.clip_data

    def _simplify_shape(self, alist, rec=0):
        """Reduce the alist dimension if needed"""
        if rec != 0:
            if len(alist) == 1:
                return alist[-1]
            return alist
        if len(alist) == 1:
            return self._simplify_shape(alist[-1], 1)
        return [self._simplify_shape(al, 1) for al in alist]

    def _get_table_data(self):
        """Return clipboard processed as data"""
        data = self._simplify_shape(self.table_widget.get_data())
        if self.table_widget.array_btn.isChecked():
            return array(data)
        elif pd and self.table_widget.df_btn.isChecked():
            info = self.table_widget.pd_info
            buf = io.StringIO(self.table_widget.pd_text)
            return pd.read_csv(buf, **info)
        return data

    def _get_plain_text(self):
        """Return clipboard as text"""
        return self.text_widget.text_editor.toPlainText()

    @Slot()
    def process(self):
        """Process the data from clipboard"""
        var_name = self.name_edt.text()
        try:
            self.var_name = str(var_name)
        except UnicodeEncodeError:
            self.var_name = str(var_name)
        if self.text_widget.get_as_data():
            self.clip_data = self._get_table_data()
        elif self.text_widget.get_as_code():
            self.clip_data = try_to_eval(str(self._get_plain_text()))
        else:
            self.clip_data = str(self._get_plain_text())
        self.accept()
